"""
Plot figures relevant to :class:`microstructpy.geometry.Rectangle`

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import collections

import microstructpy as msp

def main():
    # Plot breakdown
    breakdown(2.5, 1, 0.3, (9, 4))  # rectangle_001.png - rect breakdown
    breakdown(1, 1, 0.2, (4, 4))  # rectangle_001.png - square breakdown


def breakdown(length, width, x1, figsize):
    r = msp.geometry.Rectangle(length=length, width=width)
    approx = r.approximate(x1=x1)

    # Plot rectangle
    plt.figure(figsize=figsize)
    r.plot(edgecolor='k', facecolor='none', lw=3)

    # Plot breakdown
    t = np.linspace(0, 2 * np.pi)
    xp = np.cos(t)
    yp = np.sin(t)
    for x, y, radius in approx:
        plt.plot(x + radius * xp, y + radius * yp, 'b')

    # Format Axes
    xtick = np.unique([circ[0] for circ in approx])
    ytick = np.unique([circ[1] for circ in approx])
    plt.xticks(xtick)
    plt.yticks(ytick)

    plt.axis('scaled')
    plt.grid(True, linestyle=':')


if __name__ == '__main__':
    plt.rc('savefig', dpi=300, bbox='tight', pad_inches=0)
    main()


"""Verification

This module contains functions related to mesh verification.

"""

# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #

from __future__ import division
from __future__ import print_function

import copy
import os

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

import microstructpy.geometry
from microstructpy import _misc
from microstructpy import seeding

__author__ = 'Kenneth (Kip) Hart'

styles_vec = ['o', '^', 's', 'D', 'v', '*']

hist_class = scipy.stats._continuous_distns.rv_histogram
discr_rvs = scipy.stats._discrete_distns._distn_names
conti_rvs = scipy.stats._continuous_distns._distn_names

ori_deg_kws = ['orientation', 'angle', 'angle_deg']
ori_rad_kws = ['angle_rad']


# --------------------------------------------------------------------------- #
#                                                                             #
# Volume Fractions                                                            #
#                                                                             #
# --------------------------------------------------------------------------- #
def volume_fractions(poly_mesh, n_phases):
    """Verify volume fractions

    This function computes the volume fractions of each phase in the output
    mesh. It does so by summing the volumes of the cells in the polygonal
    mesh.

    Args:
        poly_mesh (PolyMesh): The polygonal/polyhedral mesh.
        n_phases (int): Number of phases.

    Returns:
        numpy.ndarray: Volume fractions of each phase in the poly mesh.

    """
    region_phases = np.array(poly_mesh.phase_numbers, dtype='int')
    region_volumes = np.array(poly_mesh.volumes, dtype='float')
    vol_list = np.full(n_phases, 0, dtype='float')
    for i in range(n_phases):
        mask = np.isclose(region_phases, i)
        vol_list[i] = np.sum(region_volumes[mask])

    v_total = np.sum(vol_list)
    if v_total == 0:
        return vol_list
    return vol_list / v_total


def write_volume_fractions(vol_fracs, phases, filename='volume_fractions.txt'):
    """Write volume fractions to a file

    Write the volume fractions verification out to a file.
    The output columns are:

        1. Phase number
        2. Phase name
        3. Input relative volume (average, if distributed)
        4. Output relative volume
        5. Input volume fraction (average, if distributed)
        6. Output volume fraction

    The first three lines of the output file are headings.

    Args:
        vol_fracs (list or numpy.ndarray): Volume fractions of the output mesh.
        phases (list): List of phase dictionaries.
        filename (str): *(optional)* Name of file to write.
            Defaults to ``volume_fractions.txt``.

    Returns:
        none, prints formatted volume fraction verification table to file

    """
    # Sample input volume distributions
    n_inp = 5000
    vol_trials = np.full((n_inp, len(phases)), -1, dtype='float')
    for i, phase in enumerate(phases):
        mask = vol_trials[:, i] < 0
        while np.any(mask):
            new_vals = _safe_rvs(phase.get('fraction', 1), np.sum(mask))
            vol_trials[mask, i] = new_vals
            mask = vol_trials[:, i] < 0

    # Determine input and output relative volumes
    vols_exp = np.array([_safe_mean(p.get('fraction', 1)) for p in phases])
    vols_act = np.sum(vols_exp) * np.array(vol_fracs)

    # Determine input volume fractions
    vol_total = np.sum(vol_trials, axis=1)

    k = 1 / vol_total
    frac_trials = vol_trials * k.reshape(-1, 1)
    vf_exp = np.mean(frac_trials, axis=0)

    # Create table
    hdr1 = ['',
            '',
            'Input (Avg.)',
            'Output',
            'Input (Avg.)',
            'Output']
    hdr2 = ['#',
            'Name',
            'Relative Volume',
            'Relative Volume',
            'Volume Fraction',
            'Volume Fraction']

    rows = [hdr1, hdr2]
    for i, phase in enumerate(phases):
        name = phase.get('name', 'Material ' + str(i + 1))
        row = [i,
               name,
               vols_exp[i],
               vols_act[i],
               vf_exp[i],
               vol_fracs[i]]
        rows.append(row)

    # Write table
    file_dir = os.path.dirname(filename)
    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    with open(filename, 'w') as file:
        file.write(_fixed_width_table(rows, [0, 2, 3, 4, 5]))


def plot_volume_fractions(vol_fracs, phases, filename='volume_fractions.png'):
    """Plot volume fraction verification

    This function creates a bar chart comparing the input and output volume
    fractions. If the input volume fraction is distributed, the top of the
    bar will be a curve representing the CDF of the distribution.

    Args:
        vol_fracs (list or numpy.ndarray): Output volume fractions.
        phases (list): List of phase dictionaries
        filename (str or list): *(optional)* Filename(s) to save the plot.
            Defaults to ``volume_fractions.png``.

    Returns:
        none, writes plot to file.
    """
    # Sample input volume distributions
    n_inp = 5000
    vol_trials = np.full((n_inp, len(phases)), -1, dtype='float')
    for i, phase in enumerate(phases):
        mask = vol_trials[:, i] < 0
        while np.any(mask):
            new_vals = _safe_rvs(phase.get('fraction', 1), np.sum(mask))
            vol_trials[mask, i] = new_vals
            mask = vol_trials[:, i] < 0

    # Determine input volume fractions
    vol_total = np.sum(vol_trials, axis=1)
    k = 1 / vol_total
    frac_trials = vol_trials * k.reshape(-1, 1)

    # Initialize plot
    plt.clf()
    total_w = 0.8
    bar_w = 0.5 * total_w
    n_plt = 201
    q_plt = np.linspace(0, 1, n_plt)
    x_cens = np.arange(len(phases))

    # Plot input volume fraction bars
    for i, phase in enumerate(phases):
        x_plt = x_cens[i] - bar_w + bar_w * q_plt
        y_plt = np.quantile(frac_trials[:, i], q_plt)

        x_plt = np.concatenate((x_plt, [x_plt[-1], x_plt[0]]))
        y_plt = np.concatenate((y_plt, [0, 0]))

        if i == 0:
            plt.fill(x_plt, y_plt, 'white', edgecolor='k', label='Input',
                     zorder=3)
        else:
            plt.fill(x_plt, y_plt, 'white', edgecolor='k', zorder=3)

    # Plot output volume fraction bars
    plt.bar(x_cens + 0.5 * bar_w, vol_fracs, width=bar_w, label='Actual',
            color='gray', edgecolor='black', zorder=4)

    # Format plot
    xticks = range(len(phases))
    xlbls = [p.get('name', 'Material ' + str(i + 1)) for i, p in
             enumerate(phases)]
    plt.xticks(xticks, xlbls)
    plt.gca().set_ylim(bottom=0)
    plt.grid(axis='y')

    plt.legend(loc='lower center', ncol=2, bbox_to_anchor=(0.5, -0.155),
               fancybox=True)
    plt.ylabel('Volume Fraction')
    plt.title('Volume Fraction Verification')

    if not isinstance(filename, list):
        filename = [filename]

    for fname in filename:
        write_dir = os.path.dirname(fname)
        if not os.path.exists(write_dir):
            os.makedirs(write_dir)
        plt.savefig(fname)


# --------------------------------------------------------------------------- #
#                                                                             #
# Seeds of Best Fit                                                           #
#                                                                             #
# --------------------------------------------------------------------------- #
def seeds_of_best_fit(seeds, phases, pmesh, tmesh):
    """Calculate seed geometries of best fit

    This function computes the seeds of best fit for the resultant polygonal
    and triangular meshes. It calls the the ``best_fit`` function of each
    seed's geometry, then copies the other seed attributes to create a new
    :class:`.SeedList`.

    The points on the faces of the grains are used to determine a fit geometry.
    Points on the exterior of the domain are not used since they would alter
    the shape of the best fit seed.

    Args:
        seeds (SeedList): List of seed geometries.
        phases (list): List of material phases. See :ref:`phase_dict_guide`
            for more information on formatting.
        pmesh (PolyMesh): Resultant polygonal/polyhedral mesh.
        tmesh (TriMesh): Resultant triangular/tetrahedral mesh.

    Returns:
        SeedList: List of seeds of best fit.

    .. note:

        If the geometry of best fit for the seed is ill-conditioned (e.g.
        negative axes for an ellipse), the seed geometry is set to None.
    """
    poly_pts = np.array(pmesh.points, dtype='float')
    poly_neighs = np.array(pmesh.facet_neighbors, dtype='int')
    tri_parents = np.array(tmesh.facet_attributes, dtype='int')
    tri_facets = np.array(tmesh.facets, dtype='int')
    tri_pts = np.array(tmesh.points, dtype='float')
    seed_nums = np.array(pmesh.seed_numbers, dtype='int')

    poly_neighs[poly_neighs < 0] = -1
    tri_facet_neigh_cells = poly_neighs[tri_parents]

    tri_facet_neigh_seeds = seed_nums[tri_facet_neigh_cells]
    tri_facet_neigh_seeds[tri_facet_neigh_cells < 0] = -1
    poly_facet_neigh_seeds = seed_nums[poly_neighs]
    poly_facet_neigh_seeds[poly_neighs < 0] = -1

    tri_facet_is_ext = np.min(tri_facet_neigh_seeds, axis=-1) < 0
    poly_facet_is_ext = np.min(poly_facet_neigh_seeds, axis=-1) < 0

    n_dim = seeds[0].geometry.n_dim
    fit_seeds = []
    for i, seed in enumerate(seeds):
        p = seed.phase
        mat_type = phases[p].get('material_type', 'crystalline')
        if mat_type in _misc.kw_solid:
            mask = np.any(tri_facet_neigh_seeds == i, axis=-1)
            mask &= ~tri_facet_is_ext
            seed_facets = tri_facets[mask]
            kps = np.unique(seed_facets)

            if len(kps) <= n_dim:
                mask = np.any(tri_facet_neigh_seeds == i, axis=-1)
                seed_facets = tri_facets[mask]
                kps = np.unique(seed_facets)

            seed_pts = tri_pts[kps]

        if (mat_type not in _misc.kw_solid) or (len(kps) <= n_dim):
            mask = np.any(poly_facet_neigh_seeds == i, axis=-1)
            mask &= ~poly_facet_is_ext
            seed_facets = [f for f, m in zip(pmesh.facets, mask) if m]
            kps = np.unique([kp for f in seed_facets for kp in f])

            if len(kps) <= n_dim:
                mask = np.any(poly_facet_neigh_seeds == i, axis=-1)
                seed_facets = [f for f, m in zip(pmesh.facets, mask) if m]
                kps = np.unique([kp for f in seed_facets for kp in f])
            seed_pts = poly_pts[kps.astype('int')]

        try:
            fit_geom = seed.geometry.best_fit(seed_pts)
        except ValueError:
            fit_geom = None
        fit_seed = copy.deepcopy(seed)

        # Check for ill-conditioned seeds of best fit
        if _ill_conditioned(fit_geom):
            fit_seed.geometry = None
        else:
            fit_seed.geometry = fit_geom

        fit_seeds.append(fit_seed)
    return seeding.SeedList(fit_seeds)


def _ill_conditioned(geom):
    if isinstance(geom, microstructpy.geometry.Ellipse):
        return min(geom.axes) <= 0
    return False


# --------------------------------------------------------------------------- #
#                                                                             #
# Compare to Phases                                                           #
#                                                                             #
# --------------------------------------------------------------------------- #
def plot_distributions(seeds, phases, dirname='.', ext='png', poly_mesh=None,
                       verif_mask=None):
    """Plot comparison between input and output distributions

    This function takes seeds and compares them against the input phases.
    A polygon mesh can be included for cases where grains are given an
    area or volume distribution, rather than size/shape/etc.

    This function creates both PDF and CDF plots.

    Args:
        seeds (SeedList): List of seeds to compare.
        phases (list): List of phase dictionaries.
        dirname (str): *(optional)* Plot output directory.
            Defaults to ``.``.
        ext (str or list): *(optional)* File extension(s) of the output plots.
            Defaults to ``'png'``.
        poly_mesh (PolyMesh): *(optional)* Polygonal mesh, useful for phases
            with an area or volume distribution.

    Returns:
        none, creates plot files.
    """
    if verif_mask is None:
        verif_mask = np.full(len(seeds), True)

    # Get geometry of seeds
    comp_phases = _phase_values(seeds, phases, poly_mesh, verif_mask)

    # Determine all geometry keywords used
    kws = set([kw for phase in comp_phases for kw in phase.keys()])
    kws -= set(_misc.gen_kws)

    # Create Plots
    n_dim = len(seeds[0].position)
    for kw in sorted(kws):
        if (kw in _misc.ori_kws) and (n_dim == 3):
            continue

        # PDF plot
        plt.clf()
        line_colors = []
        line_labels = []
        ymax = 0
        for i, phase in enumerate(phases):
            if (kw not in comp_phases[i]) or (kw not in phase):
                continue

            # Get input PDF
            ymax_inp_pdf = _plot_inp_pdf(kw, i, phase)
            ymax = max(ymax, ymax_inp_pdf)

            # Get output PDF
            comp_phase = comp_phases[i]
            ymax_out_pdf, lcs, lls = _plot_out_pdf(kw, i, phase, comp_phase)
            ymax = max(ymax, ymax_out_pdf)
            line_colors.extend(lcs)
            line_labels.extend(lls)

        # Format PDF plot
        color_legend = plt.legend(line_colors, line_labels, loc=4)

        style_labels = ['Input', 'Actual']
        dashed_line = Line2D([0], [0], color='k', ls=':')
        solid_line = Line2D([0], [0], color='k', ls='-')
        style_lines = [dashed_line, solid_line]
        style_labels = ['Input', 'Actual']
        style_legend = plt.legend(style_lines, style_labels, loc=2)

        plt.gca().add_artist(style_legend)
        plt.gca().add_artist(color_legend)

        plt.grid(True)
        xlbl = ' '.join([s.capitalize() for s in kw.split('_')])
        xlbl = xlbl.replace('Rad', '(radians)').replace('Deg', '(degrees)')
        xlbl = xlbl.replace('Orientation', 'Orientation (degrees)')
        plt.xlabel(xlbl)
        plt.ylabel('Probability Density Function')

        plt.ylim([0, 1.1 * ymax])

        plt.title('PDF Comparison')

        # Save PDF plot
        if not isinstance(ext, list):
            ext = [ext]

        basename = kw + '_pdf.'
        filepath = os.path.join(dirname, basename)
        for extension in ext:
            filename = filepath + extension
            plt.savefig(filename)
        plt.close()

        # CDF Plot
        plt.clf()
        line_colors = []
        line_labels = []
        for i, phase in enumerate(phases):
            if (kw not in comp_phases[i]) or (kw not in phase):
                continue

            # Get input CDF
            _plot_inp_cdf(kw, i, phase)

            # Get output PDF
            comp_phase = comp_phases[i]
            lcs, lls = _plot_out_cdf(kw, i, phase, comp_phase)
            line_colors.extend(lcs)
            line_labels.extend(lls)

        # Format CDF plot
        color_legend = plt.legend(line_colors, line_labels, loc=4)
        style_legend = plt.legend(style_lines, style_labels, loc=2)

        plt.gca().add_artist(style_legend)
        plt.gca().add_artist(color_legend)

        plt.grid(True)
        xlbl = ' '.join([s.capitalize() for s in kw.split('_')])
        xlbl = xlbl.replace('Rad', '(radians)').replace('Deg', '(degrees)')
        xlbl = xlbl.replace('Orientation', 'Orientation (degrees)')
        plt.xlabel(xlbl)
        plt.ylabel('Cumulative Distribution Function')

        plt.ylim([0, 1])

        plt.title('CDF Comparison')

        # Save PDF plot
        if not isinstance(ext, list):
            ext = [ext]

        basename = kw + '_cdf.'
        filepath = os.path.join(dirname, basename)
        for extension in ext:
            filename = filepath + extension
            plt.savefig(filename)
        plt.close()


def _plot_inp_pdf(kw, i, phase):
    ymax = 0
    inp_dist = phase[kw]
    color = phase.get('color', 'C' + str(i % 10))
    if kw in ori_deg_kws and phase[kw] == 'random':
        x_plt = [0, 360]
        y_plt = [1 / 360, 1 / 360]
        plt.plot(x_plt, y_plt, color=color, ls=':')
        ymax = 1 / 360

    elif kw in ori_rad_kws and phase[kw] == 'random':
        x_plt = [0, 2 * np.pi]
        y_plt = [0.5 / np.pi, 0.5 / np.pi]
        plt.plot(x_plt, y_plt, color=color, ls=':')
        ymax = y_plt[0]

    elif phase[kw] == 'random':
        e_str = 'Cannot create PDF for random setting'
        e_str += ' of keyword <' + str(kw) + '>'
        raise NotImplementedError(e_str)

    elif kw == 'orientation':
        ct = inp_dist[0][0]
        st = inp_dist[1][0]
        inp_deg = np.rad2deg(np.arctan2(st, ct))
        plt.plot([inp_deg, inp_deg], [0, 1e12], color=color, ls=':')

    elif isinstance(inp_dist, list):
        for j, dist in enumerate(inp_dist):
            try:
                lb = dist.ppf(1e-3)
                ub = dist.ppf(1 - 1e-3)
                x_plt = np.linspace(lb, ub, 51)
                y_plt = dist.pdf(x_plt)
                ymax = max(ymax, np.max(y_plt))
            except AttributeError:
                x_plt = [dist, dist]
                y_plt = [0, 1e12]
            m = styles_vec[j % len(styles_vec)]
            plt.plot(x_plt, y_plt, color=color, ls=':', marker=m)

    else:
        try:
            lb = inp_dist.ppf(1e-3)
            ub = inp_dist.ppf(1 - 1e-3)
            x_plt = np.linspace(lb, ub, 51)
            y_plt = inp_dist.pdf(x_plt)
            ymax = np.max(y_plt)
        except AttributeError:
            x_plt = [inp_dist, inp_dist]
            y_plt = [0, 1e12]

        plt.plot(x_plt, y_plt, color=color, ls=':')
    return ymax


def _plot_inp_cdf(kw, i, phase):
    quants = np.linspace(0, 1, 501)[1:-1]
    quants_lr = np.linspace(0, 1, 21)[1:-1]

    inp_dist = phase[kw]
    color = phase.get('color', 'C' + str(i % 10))
    if kw in ori_deg_kws and phase[kw] == 'random':
        x_plt = [0, 360]
        y_plt = [0, 1]
        plt.plot(x_plt, y_plt, color=color, ls=':')
    elif kw in ori_rad_kws and phase[kw] == 'random':
        x_plt = [0, 2 * np.pi]
        y_plt = [0, 1]
        plt.plot(x_plt, y_plt, color=color, ls=':')

    elif phase[kw] == 'random':
        e_str = 'Cannot create CDF for random setting'
        e_str += ' of keyword <' + str(kw) + '>'
        raise NotImplementedError(e_str)

    elif kw == 'orientation':
        ct = inp_dist[0][0]
        st = inp_dist[1][0]
        inp_deg = np.rad2deg(np.arctan2(st, ct))
        plt.plot([inp_deg, inp_deg], [0, 1], color=color, ls=':')

    elif isinstance(inp_dist, list):
        for j, dist in enumerate(inp_dist):
            try:
                x_plt = dist.ppf(quants)
                y_plt = quants

                x_plt_lr = dist.ppf(quants_lr)
                y_plt_lr = quants_lr
            except AttributeError:
                x_plt = np.full(11, dist)
                y_plt = np.linspace(0, 1, 11)

                x_plt_lr = x_plt
                y_plt_lr = y_plt
            m = styles_vec[j % len(styles_vec)]
            plt.plot(x_plt, y_plt, color=color, ls=':')
            plt.plot(x_plt_lr, y_plt_lr, color=color, marker=m)

    else:
        try:
            x_plt = inp_dist.ppf(quants)
            y_plt = quants
        except AttributeError:
            x_plt = [inp_dist, inp_dist]
            y_plt = [0, 1]

        plt.plot(x_plt, y_plt, color=color, ls=':')


def _plot_out_pdf(kw, i, phase, comp_phase):
    line_colors = []
    line_labels = []
    ymax = 0

    inp_dist = phase[kw]
    comp_vals = np.array([v for v in comp_phase[kw] if v is not None])
    color = phase.get('color', 'C' + str(i % 10))
    name = str(phase.get('name', 'Material ' + str(i + 1)))
    if kw == 'orientation':
        comp_ct = comp_vals[:, 0, 0]
        comp_st = comp_vals[:, 1, 0]
        comp_vals = np.rad2deg(np.arctan2(comp_st, comp_ct))

        ys, xbs, _ = plt.hist(comp_vals, density=True, histtype='step',
                              color=color)
        ymax = max(ys)

        line_colors.append(Line2D([0], [0], color=color))
        line_labels.append(name)

    elif isinstance(inp_dist, list):
        for j, vals in enumerate(comp_vals):
            ys, xbs, _ = plt.hist(vals, density=True, histtype='step',
                                  color=color)
            ymax = max(ymax, np.max(ys))
            xs = 0.5 * (xbs[:-1] + xbs[1:])

            m = styles_vec[j % len(styles_vec)]
            plt.plot(xs, ys, color=color, marker=m)

            line_colors.append(Line2D([0], [0], color=color, marker=m))
            line_labels.append(name + '[' + str(j) + ']')

    else:
        ys, xbs, _ = plt.hist(comp_vals, bins=20, density=True,
                              histtype='step', color=color)
        ymax = max(ymax, np.max(ys))

        line_colors.append(Line2D([0], [0], color=color))
        line_labels.append(name)

    return ymax, line_colors, line_labels


def _plot_out_cdf(kw, i, phase, comp_phase):
    quants = np.linspace(0, 1, 501)

    line_colors = []
    line_labels = []

    inp_dist = phase[kw]
    comp_vals = np.array([v for v in comp_phase[kw] if v is not None])
    color = phase.get('color', 'C' + str(i % 10))
    name = str(phase.get('name', 'Material ' + str(i + 1)))
    if kw == 'orientation':
        comp_ct = comp_vals[:, 0, 0]
        comp_st = comp_vals[:, 1, 0]
        comp_vals = np.rad2deg(np.arctan2(comp_st, comp_ct))

        x_plt = np.quantile(comp_vals, quants)
        y_plt = quants

        plt.plot(x_plt, y_plt, color=color)
        line_colors.append(Line2D([0], [0], color=color))
        line_labels.append(name)

    elif isinstance(inp_dist, list):
        for j, vals in enumerate(comp_vals):

            x_plt = np.quantile(vals, quants)
            y_plt = quants

            m = styles_vec[j % len(styles_vec)]
            plt.plot(x_plt, y_plt, color=color, marker=m)

            line_colors.append(Line2D([0], [0], color=color, marker=m))
            line_labels.append(name + '[' + str(j) + ']')

    else:
        x_plt = np.quantile(comp_vals, quants)
        y_plt = quants

        plt.plot(x_plt, y_plt, color=color)
        line_colors.append(Line2D([0], [0], color=color))
        line_labels.append(name)

    return line_colors, line_labels


# --------------------------------------------------------------------------- #
#                                                                             #
# Maximum Likelihood Estimators for Phases                                    #
#                                                                             #
# --------------------------------------------------------------------------- #
def mle_phases(seeds, phases, poly_mesh=None, verif_mask=None):
    """Get maximum likelihood estimators (MLEs) for phases

    This function finds distributions in the list of phases and computes the
    MLE parameters for those distributions. The returned value is a list of
    phases with the same length and dictionary keywords, except the
    distributions are replaced with MLE distributions (based on the seeds).
    Constant values are replaced with the mean of the seed values.

    Note that the directional statistics are not used - so the results for
    orientation angles and matrices are unreliable.

    Also note that SciPy currently does not support MLEs for discrete random
    variables. Any discrete distributions will be given a histogram output.

    .. note::

        Directional statistics are not used and as such the results for
        orientation angles and matrices are unreliable. The only exception
        is normally distributed orientation angles.

    Args:
        seeds (SeedList): List of seeds.
        phases (list): List of input phase dictionaries.
        poly_mesh (PolyMesh): *(optional)* Polygonal/polyhedral mesh.
        verif_mask (list or numpy.ndarray): *(optional)* Mask for which
            seeds to include in the MLE parameter calculation. Default is
            True for all seeds.
    """
    circ_highs = {'angle': 360, 'angle_deg': 360, 'angle_rad': 2 * np.pi}

    if verif_mask is None:
        verif_mask = np.full(len(seeds), True)

    # Get geometry of seeds
    comp_phases = _phase_values(seeds, phases, poly_mesh, verif_mask)

    # Get MLEs
    param_phases = []
    for comp_phase, phase in zip(comp_phases, phases):
        param_phase = {}
        for kw in comp_phase:
            comp_vals = np.array([v for v in comp_phase[kw] if v is not None])
            inp_dist = phase[kw]
            can_circ = 'angle' in kw and hasattr(inp_dist, 'dist')
            if can_circ:
                can_circ &= inp_dist.dist.name == 'norm'

            if can_circ:
                high = circ_highs[kw]
                circ_mean = scipy.stats.circmean(comp_vals, high=high)
                circ_scl = scipy.stats.circstd(comp_vals, high=high)
                dist = scipy.stats.norm(loc=circ_mean, scale=circ_scl)
            else:
                dist = _mle_dist(comp_vals, inp_dist)
            param_phase[kw] = dist
        param_phases.append(param_phase)
    return param_phases


def write_mle_phases(inp_phases, out_phases, filename='mles.txt'):
    """Write MLE parameters in a table

    This function writes out a text file containing the input parameters and
    maximum likelihood estimators (MLEs) for the outputs.

    Args:
        inp_phases (list): List of input phase dictionaries.
        out_phases (list): List of output phase dictionaries.
        filename (str): *(optional)* Filename of the output table.
            Defaults to ``mles.txt``.

    Returns:
        none, writes file.

    """
    # Create table rows as dictionaries
    rows_dict = []

    assert len(inp_phases) == len(out_phases)
    for i in range(len(inp_phases)):
        inp_phase = inp_phases[i]
        out_phase = out_phases[i]

        name = inp_phase.get('name', 'Material ' + str(i + 1))

        kws = set(out_phase.keys()) - set(_misc.gen_kws)
        for kw in sorted(kws):
            inp_dist = inp_phase[kw]
            out_dist = out_phase[kw]

            if kw == 'orientation':
                row_dict = {'i': i, 'name': name, 'kw': kw}
                rows_dict.append(row_dict)
                continue
            if isinstance(inp_dist, list):
                for j in range(len(inp_dist)):
                    row_dict = {'i': i, 'name': name, 'kw': kw + '[' + j + ']'}
                    inp_dict = _dist_dict(inp_dist[j])
                    out_dict = _dist_dict(out_dist[j])
                    for key in inp_dict:
                        row_dict[key + '_inp'] = inp_dict[key]
                        row_dict[key + '_out'] = out_dict[key]
                    rows_dict.append(row_dict)
            else:
                row_dict = {'i': i, 'name': name, 'kw': kw}
                inp_dict = _dist_dict(inp_dist)
                out_dict = _dist_dict(out_dist)
                for key in inp_dict:
                    row_dict[key + '_inp'] = inp_dict[key]
                    row_dict[key + '_out'] = out_dict[key]
                rows_dict.append(row_dict)

    # Determine headings / order of dictionaries
    hdg_kws = set([kw for row in rows_dict for kw in row.keys()])
    init_kws = ['i', 'name', 'kw']
    loc_scale_kws = ['loc_inp', 'loc_out', 'scale_inp', 'scale_out']

    all_kws = [kw for kw in init_kws]
    for kw in sorted(list(hdg_kws)):
        if kw in init_kws or kw in loc_scale_kws:
            continue
        all_kws.append(kw)

    for kw in loc_scale_kws:
        if kw in hdg_kws:
            all_kws.append(kw)

    # create list rows
    rows_list = [[d.get(kw, '') for kw in all_kws] for d in rows_dict]

    # create headers
    hdr = _mle_hdr(all_kws)
    table = hdr + rows_list

    rjust = [i for i, kw in enumerate(all_kws) if kw not in ['name', 'kw']]

    # Save table
    out_dir = os.path.dirname(filename)
    if out_dir and not os.path.exists(out_dir):
        os.makedirs(out_dir)

    with open(filename, 'w') as file:
        file.write(_fixed_width_table(table, rjust))


def _mle_hdr(all_kws):
    hdr1 = []
    hdr2 = []
    for kw in all_kws:
        if kw == 'i':
            h1 = ''
            h2 = '#'
        elif kw == 'name':
            h1 = ''
            h2 = 'Name'
        elif kw == 'kw':
            h1 = ''
            h2 = 'Parameter'
        elif kw.endswith('_inp'):
            h1 = 'Input'
            h2 = kw.rstrip('_inp')
        elif kw.endswith('_out'):
            h1 = 'Output'
            h2 = kw.rstrip('_out')
        else:
            raise ValueError('Cannot creating heading for keyword ' + str(kw))
        hdr1.append(h1)
        hdr2.append(h2)
    return [hdr1, hdr2]


# --------------------------------------------------------------------------- #
#                                                                             #
# Phase Error Statistics                                                      #
#                                                                             #
# --------------------------------------------------------------------------- #
def error_stats(fit_seeds, seeds, phases, poly_mesh=None, verif_mask=None):
    """Error statistics for seeds

    This function creates a dictionary of error statistics for each of the
    input distributions in the phases.

    Args:
        fit_seeds (SeedList): List of seeds of best fit.
        seeds (SeedList): List of seeds.
        phases (list): List of input phase dictionaries.
        poly_mesh (PolyMesh): *(optional)* Polygonal/polyhedral mesh.
        verif_mask (list or numpy.ndarray): *(optional)* Mask for seeds to
            be included in the analysis. Defaults to all True.

    Returns:
        list: List with the same size and dictionary keywords as phases,
        but with error statistics dictionaries in each entry.

    """

    if verif_mask is None:
        verif_mask = np.full(len(seeds), True)

    # Organize the geometry values
    init_phases = _phase_values(seeds, phases, verif_mask=verif_mask)
    outp_phases = _phase_values(fit_seeds, phases, poly_mesh, verif_mask)

    err_phases = []
    for i in range(len(phases)):
        i_phase = init_phases[i]
        o_phase = outp_phases[i]
        phase = phases[i]
        for kw in phase:
            if kw in ('angle', 'angle_deg') and phase[kw] == 'random':
                phase[kw] = scipy.stats.uniform(loc=0, scale=360)
            if kw == 'angle_rad':
                phase[kw] = scipy.stats.uniform(loc=0, scale=2 * np.pi)

        err_io = {kw: _kw_errs(i_phase[kw], o_phase[kw]) for kw in i_phase}
        err_po = {kw: _kw_stats(phase[kw], o_phase[kw]) for kw in o_phase}

        err_phase = {}
        for kw in i_phase:
            if kw == 'orientation':
                err_phase[kw] = {}
                continue
            val = err_io[kw].copy()
            val.update(err_po[kw])
            err_phase[kw] = val

        err_phases.append(err_phase)
    return err_phases


def _kw_errs(y_exp, y_act):
    if np.array(y_exp).ndim > 1:
        return [_kw_errs(*tup) for tup in zip(y_exp, y_act)]

    errs = {}

    mask = np.array([y_a is not None for y_a in y_act])
    if not np.any(mask):
        return errs

    y_expect = np.array(y_exp)[mask]
    y_actual = np.array([y_a for y_a in y_act if y_a is not None])

    r = y_actual - y_expect

    # MAE
    mae = np.mean(np.abs(r))
    errs['mae'] = mae

    # MSE
    mse = np.mean(r * r)
    errs['mse'] = mse

    # RMSE
    rmse = np.sqrt(mse)
    errs['rmse'] = rmse

    # R^2
    coeff_det = _r2(y_actual, y_expect)
    errs['R^2'] = coeff_det

    # Max Error
    inf_norm = np.max(np.abs(r))
    errs['inf_norm'] = inf_norm

    return errs


def _r2(y_act, y_exp):
    r = y_act - y_exp
    mse = np.mean(r * r)
    y_bar = np.mean(y_act)
    r_ybar = y_act - y_bar
    mse_baseline = np.mean(r_ybar * r_ybar)

    coeff_det = 1 - (mse / mse_baseline)
    return coeff_det


def _kw_stats(dist_exp, y_act):
    if isinstance(dist_exp, list):
        return [_kw_stats(*tup) for tup in zip(dist_exp, y_act)]

    stats = {}
    y_actual = np.array([y_a for y_a in y_act if y_a is not None])
    if len(y_actual) == 0:
        return stats

    y_pred = _safe_rvs(dist_exp, 5000)

    # Wasserstein Distance
    wass = scipy.stats.wasserstein_distance(y_actual, y_pred)
    stats['wasserstein_distance'] = wass

    # Energy Distance
    e_dist = scipy.stats.energy_distance(y_actual, y_pred)
    stats['energy_distance'] = e_dist

    # K-S Test
    if hasattr(dist_exp, 'cdf'):
        cdf_func = dist_exp.cdf
    else:
        def cdf_func(x):
            return (x > dist_exp).astype('float')

    ks_stat, ks_p = scipy.stats.kstest(y_actual, cdf_func)
    stats['ks_statistic'] = ks_stat
    stats['ks_p_value'] = ks_p

    # T-test
    t_stat, t_p = scipy.stats.ttest_ind(y_actual, y_pred)
    stats['t_stat'] = t_stat
    stats['t_p_value'] = t_p

    return stats


def write_error_stats(errs, phases, filename='error_stats.txt'):
    """Write error statistics to file

    This function takes previously computed error statistics and writes them
    to a human-readable text file.

    Args:
        errs (list): List of error statistics for each input phase parameter.
            Organized the same as ``phases``.
        phases (list): List of input phases. See :ref:`phase_dict_guide` for
            more details.
        filename (str): *(optional)* The name of the file to contain the
            error statistics. Defaults to ``error_stats.txt``.
    """
    # Create table rows as dictionaries
    rows_dict = []

    assert len(errs) == len(phases)
    for i in range(len(errs)):
        err_dict = errs[i]
        phase = phases[i]

        name = phase.get('name', 'Material ' + str(i + 1))

        kws = set(err_dict.keys()) - set(_misc.gen_kws)
        for kw in kws:
            err_metrics = err_dict[kw]
            inp_dist = phase[kw]
            if isinstance(inp_dist, list):
                for j in range(len(inp_dist)):
                    row_dict = {'i': i, 'name': name, 'kw': kw + '[' + j + ']'}
                    row_dict.update(err_metrics[j])
                    rows_dict.append(row_dict)
            else:
                row_dict = {'i': i, 'name': name, 'kw': kw}
                row_dict.update(err_metrics)
                rows_dict.append(row_dict)

    # Determine headings / order of dictionaries
    hdg_kws = set([kw for row in rows_dict for kw in row.keys()])
    init_kws = ['i', 'name', 'kw']

    all_kws = [kw for kw in init_kws]
    for kw in sorted(list(hdg_kws)):
        if kw not in init_kws:
            all_kws.append(kw)

    # create list rows
    rows_list = [[d.get(kw, '') for kw in all_kws] for d in rows_dict]

    # create headers
    hdr1 = []
    hdr2 = []

    hd1 = {'i': '', 'name': '', 'kw': '',
           'energy_distance': 'Energy',
           'inf_norm': 'Maximum',
           'ks_p_value': 'K-S',
           'ks_statistic': 'K-S',
           'mae': 'Mean Average',
           'mse': 'Mean Squared',
           'rmse': 'Root Mean',
           't_p_value': 't-test',
           't_stat': 't-test',
           'wasserstein_distance': 'Wasserstein'}

    hd2 = {'i': '#', 'name': 'Name', 'kw': 'Parameter',
           'energy_distance': 'Distance',
           'inf_norm': 'Absolute Error',
           'ks_p_value': 'p-value',
           'ks_statistic': 'Statistic',
           'mae': 'Error',
           'mse': 'Error',
           'rmse': 'Squared Error',
           't_p_value': 'p-value',
           't_stat': 'Statistic',
           'wasserstein_distance': 'Distance'}
    for kw in all_kws:
        hdr1.append(hd1.get(kw, ''))
        hdr2.append(hd2.get(kw, kw))

    table = [hdr1, hdr2] + rows_list

    rjust = [i for i, kw in enumerate(all_kws) if kw not in ['name', 'kw']]

    # Save table
    out_dir = os.path.dirname(filename)
    if out_dir and not os.path.exists(out_dir):
        os.makedirs(out_dir)

    with open(filename, 'w') as file:
        file.write(_fixed_width_table(table, rjust))


# --------------------------------------------------------------------------- #
#                                                                             #
# Private Functions                                                           #
#                                                                             #
# --------------------------------------------------------------------------- #
def _safe_mean(x):
    try:
        mu = x.mean()
    except AttributeError:
        mu = x
    return mu


def _safe_rvs(x, size=1):
    if isinstance(x, list):
        return np.array([_safe_rvs(xi, size) for xi in x]).T

    try:
        samples = x.rvs(size=size)
    except AttributeError:
        samples = np.full(size, x)
    return samples


def _phase_values(seeds, phases, poly_mesh=None, verif_mask=None):
    """Takes the properties of the seeds and organizes them like the phases

    """
    if verif_mask is None:
        verif_mask = np.full(len(seeds), True)

    if poly_mesh is not None:
        cell_vols = np.array(poly_mesh.volumes)
        vols = []
        for i in range(len(seeds)):
            mask = np.isclose(poly_mesh.seed_numbers, i)
            vols.append(np.sum(cell_vols[mask]))

    comp_phases = []
    for i, phase in enumerate(phases):
        comp_phase = {}
        phase_seeds = [s for f, s in zip(verif_mask, seeds)
                       if s.phase == i and f]
        for kw in phase:
            if 'rot_seq' in kw:
                continue
            if kw in _misc.gen_kws:
                continue

            try:
                if kw in ('area', 'volume') and (poly_mesh is not None):
                    vals = [v for v, s, f in zip(vols, seeds, verif_mask)
                            if s.phase == i and f]
                else:
                    vals = [_getattr(s.geometry, kw) for s in phase_seeds]

                comp_phase[kw] = vals
            except AttributeError:
                pass

        comp_phases.append(comp_phase)
    return comp_phases


def _getattr(inst, kw):
    try:
        a = getattr(inst, kw)
    except AttributeError:
        a = None
    return a


def _mle_dist(values, dist):
    if isinstance(dist, list):
        return [_mle_dist(*tup) for tup in zip(values, dist)]

    if not (hasattr(dist, 'dist') or isinstance(dist, hist_class)):
        return np.mean(values)

    if isinstance(dist, hist_class) or dist.dist.name in discr_rvs:
        hist = np.histogram(values, bins='auto', density=True)
        return scipy.stats.rv_histogram(hist)

    args = dist.dist.fit(values)
    kwargs = {'loc': args[-2], 'scale': args[-1]}

    name = dist.dist.name
    return scipy.stats.__dict__[name](*args[:-2], **kwargs)


def _dist_dict(dist):
    if not (hasattr(dist, 'dist') or isinstance(dist, hist_class)):
        return {'loc': dist}

    if isinstance(dist, hist_class) or dist.dist.name in discr_rvs:
        return {}

    args = dist.args
    kwds = dist.kwds
    shape_vals, loc, scale = dist.dist._parse_args(*args, **kwds)

    param_dict = {'loc': loc, 'scale': scale}
    if len(shape_vals) == 0:
        return param_dict

    shape_kwds = dist.dist.shapes.split(',')
    for val, kwd in zip(shape_vals, shape_kwds):
        param_dict[kwd.strip().lower()] = val
    return param_dict


def _fixed_width_table(rows, rjust_cols=[]):
    # heading row
    hr = None
    for i, row in enumerate(rows):
        if all([isinstance(col, str) for col in row]):
            hr = i

    # Convert rows to strings
    str_rows = []
    for row in rows:
        str_row = []
        for col in row:
            if isinstance(col, float):
                if col == 0 or np.abs(np.log10(np.abs(col))) < 3:
                    col_str = '{:f}'.format(col)
                else:
                    col_str = '{:e}'.format(col)
            else:
                col_str = str(col)
            str_row.append(col_str)
        str_rows.append(str_row)

    # Determine column widths
    col_widths = [max([len(s) for s in col]) for col in zip(*str_rows)]

    # Add divider
    if hr is not None:
        str_rows.insert(hr + 1, [w * '-' for w in col_widths])

    # Create table
    table = ''
    for i, row in enumerate(str_rows):
        row_strings = []
        for j, col in enumerate(row):
            w = col_widths[j]
            if j in rjust_cols and i > hr + 1:
                row_strings.append(col.rjust(w))
            else:
                row_strings.append(col.ljust(w))
        table += ' '.join(row_strings) + '\n'
    return table



# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
from __future__ import division

import ellipse as lsqel  # ellipse module in lsq-ellipse package
import numpy as np
from matplotlib import patches
from matplotlib import pyplot as plt

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Ellipse Class                                                               #
#                                                                             #
# --------------------------------------------------------------------------- #
class Ellipse(object):
    r"""A 2-D ellipse geometry.

    This class contains a 2-D ellipse. It is defined by a center point, axes
    and an orientation.

    Without any parameters, the ellipse defaults to a unit circle.

    Args:
        a (float): *(optional)* Semi-major axis of ellipse. Defaults to 1.
        b (float): *(optional)* Semi-minor axis of ellipse. Defaults to 1.
        center (list): *(optional)* The ellipse center.
            Defaults to (0, 0).
        axes (list): *(optional)* A 2-element list of semi-axes, equivalent
            to ``[a, b]``. Defaults to [1, 1].
        size (float): *(optional)* The diameter of a circle with equivalent
            area. Defaults to 1.
        aspect_ratio (float): *(optional)* The ratio of x-axis to y-axis
            length. Defaults to 1.
        angle (float): *(optional)* The counterclockwise rotation angle,
            in degrees, measured from the +x axis.
        angle_deg (float): *(optional)* The rotation angle, in degrees.
        angle_rad (float): *(optional)* The rotation angle, in radians.
        matrix (numpy.ndarray): *(optional)* The 2x2 rotation matrix.
        orientation (numpy.ndarray): *(optional)* Alias for ``matrix``.
    """
    # ----------------------------------------------------------------------- #
    # Constructor                                                             #
    # ----------------------------------------------------------------------- #
    def __init__(self, **kwargs):
        # Check Values
        for kw in ('a', 'b', 'size', 'aspect_ratio'):
            if kw in kwargs and kwargs[kw] <= 0:
                raise ValueError(kw + ' should be positive.')
        if 'axes' in kwargs:
            for i, ax in kwargs['axes']:
                if ax <= 0:
                    raise ValueError('axes[{}] should be positive'.format(i))

        for kw in ['matrix', 'orientation']:
            if kw in kwargs:
                m = np.array(kwargs[kw])
                if m.shape != (2, 2):
                    raise ValueError(kw + ' should be 2x2.')
                if not np.all(np.isclose(m * m.T, np.eye(2))):
                    raise ValueError(kw + ' should be orthonormal.')

        # position
        if 'center' in kwargs:
            self.center = kwargs['center']

        elif 'position' in kwargs:
            self.center = kwargs['position']

        else:
            self.center = (0, 0)

        # axes
        if 'area' in kwargs:
            kwargs['size'] = 2 * np.sqrt(kwargs['area'] / np.pi)
        if ('a' in kwargs) and ('b' in kwargs):
            assert kwargs['a'] > 0
            assert kwargs['b'] > 0

            self.a = kwargs['a']
            self.b = kwargs['b']

        elif 'axes' in kwargs:
            self.a, self.b = kwargs['axes']

        elif ('size' in kwargs) and ('aspect_ratio' in kwargs):
            assert kwargs['size'] > 0
            assert kwargs['aspect_ratio'] > 0

            # Derivation for converting (r, k) into (a, b)
            # Area formula: pi*a*b = pi*r^2
            # Aspect ratio: k = a/b
            #
            # Plug AR into area: k*b^2 = r^2
            # Solve for b: b = r/sqrt(k)
            # Solve for a: a = k * b

            r = 0.5 * kwargs['size']
            k = kwargs['aspect_ratio']

            self.b = r / np.sqrt(k)
            self.a = k * self.b

        elif ('a' in kwargs) and ('size' in kwargs):
            assert kwargs['a'] > 0
            assert kwargs['size'] > 0

            r = 0.5 * kwargs['size']

            self.a = kwargs['a']
            self.b = (r * r) / self.a

        elif ('b' in kwargs) and ('size' in kwargs):
            assert kwargs['b'] > 0
            assert kwargs['size'] > 0

            r = 0.5 * kwargs['size']

            self.b = kwargs['b']
            self.a = (r * r) / self.b
        elif ('a' in kwargs) and ('aspect_ratio' in kwargs):
            assert kwargs['a'] > 0
            assert kwargs['aspect_ratio'] > 0

            self.a = kwargs['a']
            self.b = self.a / kwargs['aspect_ratio']
        elif ('b' in kwargs) and ('aspect_ratio' in kwargs):
            assert kwargs['b'] > 0
            assert kwargs['aspect_ratio'] > 0

            self.b = kwargs['b']
            self.a = self.b * kwargs['aspect_ratio']

        else:
            self.a = 1
            self.b = 1

        # orientation
        if 'angle_deg' in kwargs:
            self.angle = kwargs['angle_deg']

        elif 'angle_rad' in kwargs:
            self.angle = 180 * kwargs['angle_rad'] / np.pi

        elif 'angle' in kwargs:
            self.angle = kwargs['angle']

        elif 'matrix' in kwargs:
            ct = kwargs['matrix'][0][0]
            st = kwargs['matrix'][1][0]
            self.angle = 180 * np.arctan2(st, ct) / np.pi

        elif 'orientation' in kwargs:
            ct = kwargs['orientation'][0][0]
            st = kwargs['orientation'][1][0]
            self.angle = 180 * np.arctan2(st, ct) / np.pi

        else:
            self.angle = 0

    # ----------------------------------------------------------------------- #
    # Best Fit Function                                                       #
    # ----------------------------------------------------------------------- #
    def best_fit(self, points):
        r"""Find ellipse of best fit for points

        This function computes the ellipse of best fit for a set of points.
        It calls the `least-squares-ellipse-fitting`_ package, which implements
        a published fitting algorithm in Python. [#halir]_

        The current instance of the class is used as an initial guess for
        the ellipse of best fit. Since an ellipse can be expressed multiple
        ways (e.g. rotate 90 degrees and flip the axes), this initial guess
        is used to choose from the multiple parameter sets.

        Args:
            points (list or numpy.ndarray): An Nx2 list of points to fit.

        Returns:
            .Ellipse: An instance of the class that best fits the points.

        .. _`least-squares-ellipse-fitting`: https://github.com/bdhammel/least-squares-ellipse-fitting

        .. [#halir] Halir, R., Flusser, J., "Numerically Stable Direct Least
          Squares Fitting of Ellipses," *6th International Conference in Central
          Europe on Computer Graphics and Visualization*, Vol. 98, 1998.
          (http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.1.7559&rep=rep1&type=pdf)

        """  # NOQA: E501

        pts = np.array(points)
        pt_cen = pts.mean(axis=0)
        pts -= pt_cen.reshape(1, -1)

        reg = lsqel.LsqEllipse().fit(pts)
        center, width, height, phi = reg.as_parameters()
        xc, yc = center

        xc += pt_cen[0]
        yc += pt_cen[1]

        # Find pair closest to self
        s = np.sin(phi)
        c = np.cos(phi)
        rot = np.array([[c, -s], [s, c]])

        x_ax_seed = np.array(self.matrix)[:, 0]
        x_dot, y_dot = rot.T.dot(x_ax_seed)

        if np.abs(x_dot) > np.abs(y_dot):
            if x_dot > 0:
                x_ax_fit = rot[:, 0]
            else:
                x_ax_fit = - rot[:, 0]

            a = width
            b = height
        else:
            if y_dot > 0:
                x_ax_fit = rot[:, 1]
            else:
                x_ax_fit = - rot[:, 1]

            a = height
            b = width

        ang_diff = np.arcsin(np.cross(x_ax_seed, x_ax_fit))
        ang_rad = self.angle_rad + ang_diff

        return type(self)(center=(xc, yc), a=a, b=b, angle_rad=ang_rad)

    # ----------------------------------------------------------------------- #
    # String and Representation Functions                                     #
    # ----------------------------------------------------------------------- #
    def __str__(self):
        str_str = 'center: ' + str(tuple(self.center)) + '\n'
        str_str += 'a: ' + str(self.a) + '\n'
        str_str += 'b: ' + str(self.b) + '\n'
        str_str += 'angle: ' + str(self.angle)
        return str_str

    def __repr__(self):
        repr_str = 'Ellipse('
        repr_str += 'center=' + repr(tuple(self.center))
        repr_str += ', a=' + repr(self.a) + ', b=' + repr(self.b)
        repr_str += ', angle=' + repr(self.angle)
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Size and Orientation Getters                                            #
    # ----------------------------------------------------------------------- #
    @property
    def size(self):
        """float: Diameter of equivalent area circle"""
        return 2 * np.sqrt(self.a * self.b)

    @property
    def aspect_ratio(self):
        """float: Ratio of x-axis length to y-axis length"""
        return self.a / self.b

    @property
    def axes(self):
        """ tuple: List of semi-axes."""
        return self.a, self.b

    @property
    def angle_deg(self):
        """float: Rotation angle, in degrees"""
        return self.angle

    @property
    def angle_rad(self):
        """float: Rotation angle, in radians"""
        return self.angle * np.pi / 180

    @property
    def matrix(self):
        """numpy.ndarray: Rotation matrix"""
        ct = np.cos(self.angle_rad)
        st = np.sin(self.angle_rad)
        return np.array([[ct, -st], [st, ct]])

    @property
    def orientation(self):
        """numpy.ndarray: Rotation matrix"""
        return self.matrix

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: Number of dimensions, 2"""
        return 2

    # ----------------------------------------------------------------------- #
    # Area Property                                                           #
    # ----------------------------------------------------------------------- #
    @property
    def area(self):
        r""" float: Area of ellipse, :math:`A = \pi a b`"""
        return np.pi * self.a * self.b

    @property
    def volume(self):
        r""" float: Same as :any:`microstructpy.geometry.Ellipse.area`"""
        return self.area

    # ----------------------------------------------------------------------- #
    # Expected Area                                                           #
    # ----------------------------------------------------------------------- #
    @classmethod
    def area_expectation(cls, **kwargs):
        r"""Expected value of area.

        This function computes the expected value for the area of an ellipse.
        The keyword arguments are the same as the input parameters of the
        class. The keyword values can be either constants (ints or floats) or
        distributions from the SciPy :mod:`scipy.stats` module.

        If an ellipse is specified by size, the expected value is computed
        as follows.

        .. math::

            \mathbb{E}[A] &= \frac{\pi}{4} \mathbb[S^2] \\
                          &= \frac{\pi}{4} (\mu_S^2 + \sigma_S^2)

        If the ellipse is specified by independent distributions for each
        semi-axis, the expected value is computed by:

        .. math::

            \mathbb{E}[A] = \pi\, \mathbb{E}[A B] = \pi \mu_A \mu_B

        If the ellipse is specified by the second semi-axis and the aspect
        ratio, the expected value is computed by:

        .. math::

            \mathbb{E}[A] &= \pi\, \mathbb{E}[K B^2] \\
                          &= \pi \mu_K (\mu_B^2 + \sigma_B^2)

        Finally, if the ellipse is specified by the first semi-axis and the
        aspect ratio, the expected value is computed by Monte Carlo:

        .. math::

            \mathbb{E}[A] &= \pi\, \mathbb{E}\left[\frac{A^2}{K}\right] \\
                          &\approx \frac{\pi}{n} \sum_{i=1}^n \frac{A_i}{K_i}

        where :math:`n=1000`.

        Args:
            **kwargs: Keyword arguments, see
                :class:`microstructpy.geometry.Ellipse`.

        Returns:
            float: Expected value of the area of the ellipse.

        """  # NOQA: E501
        if 'size' in kwargs:
            s_dist = kwargs['size']

            if type(s_dist) in (float, int):
                return 0.25 * np.pi * s_dist * s_dist
            return 0.25 * np.pi * s_dist.moment(2)

        if 'area' in kwargs:
            a_dist = kwargs['area']
            try:
                a_exp = a_dist.moment(1)
            except AttributeError:
                a_exp = a_dist
            return a_exp

        if ('a' in kwargs) and ('b' in kwargs):
            exp = np.pi
            for kw in ('a', 'b'):
                dist = kwargs[kw]
                if type(dist) in (float, int):
                    mu = dist
                else:
                    mu = dist.moment(1)
                exp *= mu
            return exp

        if ('b' in kwargs) and ('aspect_ratio' in kwargs):
            exp = np.pi
            try:
                exp *= kwargs['b'].moment(2)
            except AttributeError:
                exp *= kwargs['b'] * kwargs['b']

            try:
                exp *= kwargs['aspect_ratio'].moment(1)
            except AttributeError:
                exp *= kwargs['aspect_ratio']
            return exp

        if ('a' in kwargs) and ('aspect_ratio' in kwargs):
            n = 1000
            try:
                a = kwargs['a'].rvs(size=n)
            except AttributeError:
                a = np.full(n, kwargs['a'])

            try:
                k = kwargs['aspect_ratio'].rvs(size=n)
            except AttributeError:
                k = np.full(n, kwargs['aspect_ratio'])
            return np.pi * np.mean((a * a) / k)

        e_str = 'Could not calculate expected area from keywords '
        e_str += str(kwargs.keys()) + '.'
        raise KeyError(e_str)

    # ----------------------------------------------------------------------- #
    # Bounding Circles                                                        #
    # ----------------------------------------------------------------------- #
    @property
    def bound_max(self):
        """tuple: Maximum bounding circle of ellipse, (x, y, r)"""
        r = max(self.a, self.b)
        return tuple(list(self.center) + [r])

    @property
    def bound_min(self):
        """tuple: Minimum interior circle of ellipse, (x, y, r)"""
        r = min(self.a, self.b)
        return tuple(list(self.center) + [r])

    # ----------------------------------------------------------------------- #
    # Circle Approximation of Ellipse                                         #
    # ----------------------------------------------------------------------- #
    def approximate(self, x1=None):
        """Approximate ellipse with a set of circles.

        This function converts an ellipse into a set of circles.
        It implements a published algorithm by Ilin and Bernacki. [#ilin]_

        Example:
            >>> import matplotlib.pyplot as plt
            >>> import microstructpy as msp
            >>> import numpy as np
            >>> ellipse = msp.geometry.Ellipse(a=3, b=1)
            >>> approx = ellipse.approximate(0.7)
            >>> approx
            array([[ 0.        ,  0.        ,  1.        ],
                   [ 0.7       ,  0.        ,  0.96889112],
                   [ 1.38067777,  0.        ,  0.87276349],
                   [ 2.00213905,  0.        ,  0.7063497 ],
                   [ 2.5234414 ,  0.        ,  0.45169729],
                   [ 2.66666667,  0.        ,  0.33333333],
                   [-0.7       ,  0.        ,  0.96889112],
                   [-1.38067777,  0.        ,  0.87276349],
                   [-2.00213905,  0.        ,  0.7063497 ],
                   [-2.5234414 ,  0.        ,  0.45169729],
                   [-2.66666667,  0.        ,  0.33333333]])
            >>> ellipse.plot(edgecolor='k', facecolor='none', lw=3)
            >>> t = np.linspace(0, 2 * np.pi)
            >>> for x, y, r in approx:
            ...     plt.plot(x + r * np.cos(t), y + r * np.sin(t), 'b')
            >>> plt.xticks(np.unique(np.concatenate((approx[:, 0], (-3, 3)))))
            >>> plt.yticks(np.unique(np.concatenate((approx[:, 1], (-1, 1)))))
            >>> plt.axis('scaled')
            >>> plt.grid(True, linestyle=':')
            >>> plt.show()

        Executing the code above produces :numref:`f_api_ellipse_approx`.

        .. _f_api_ellipse_approx:
        .. figure:: ../../auto_examples/geometry/images/sphx_glr_plot_ellipse_001.png

            Circular approximation of ellipse, after Ilin and Bernacki.

        Args:
            x1 (float or None): *(optional)* Position of the first circle
                along the +x axis. Defaults to 0.5x the shortest semi-axis.

        Returns:
            numpy.ndarray: An Nx3 list of the (x, y, r) data of each circle
            approximating the ellipse.

        Raises:
            AssertionError: Thrown if max(a, b) < x1.

        .. [#ilin] Ilin, D.N., and Bernacki, M., "Advancing Layer Algorithm
            of Dense Ellipse Packing for Generating Statistically Equivalent
            Polygonal Structures," Granular Matter, vol. 18(3), pp. 43, 2016.

        """  # NOQA: E501
        if x1 is None:
            x1 = 0.5 * min(self.a, self.b)

        flip = self.a < self.b
        if flip:
            a = self.b
            b = self.a
        else:
            a = self.a
            b = self.b

        if a == b:
            return np.append(self.center, a).reshape(1, -1)

        a_str = 'Center of first circle, x1=' + str(x1) + ', must be less than'
        a_str += ' semi-major axis, a=' + str(a)
        assert x1 < a, a_str

        R_N = b * b / a  # Eq. 8
        x_N = a - R_N  # Eq 9

        def R_i(x):
            return b * np.sqrt(1 - x * x / (a * a - b * b))  # Eq. 6

        def y_i(x):
            ratio = x / a
            return b * np.sqrt(1 - ratio * ratio)  # Eq. 1

        circles = [(0, 0, b)]
        y_vals = [b]

        adding_circles = x1 < x_N
        if adding_circles:
            circles.append((x1, 0, R_i(x1)))
            y_vals.append(y_i(x1))

        while adding_circles:
            y_ratio = y_vals[-1] / y_vals[-2]
            x_diff = circles[-1][0] - circles[-2][0]
            x_ip1 = y_ratio * x_diff + circles[-1][0]  # Eq. 7

            adding_circles = x_ip1 < x_N
            if adding_circles:
                circle = (x_ip1, 0, R_i(x_ip1))
                circles.append(circle)
                y_vals.append(y_i(x_ip1))
        circles.append((x_N, 0, R_N))
        reflect = [(-x, y, r) for x, y, r in circles[1:]]

        all_circles = np.array(circles + reflect)
        if flip:
            all_circles[:, [0, 1]] = all_circles[:, [1, 0]]

        rot_cens = all_circles[:, :-1].dot(self.matrix.T)
        all_circles[:, :-1] = rot_cens + np.array(self.center)
        return all_circles

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the ellipse.

        This function adds a :class:`matplotlib.patches.Ellipse` patch to the
        current axes using matplotlib. The keyword arguments are passed to
        the patch.

        Args:
            **kwargs (dict): Keyword arguments for matplotlib.

        """  # NOQA: E501
        p = patches.Ellipse(self.center, 2 * self.a, 2 * self.b,
                            self.angle_deg, **kwargs)
        plt.gca().add_patch(p)

    # ----------------------------------------------------------------------- #
    # Limits                                                                  #
    # ----------------------------------------------------------------------- #
    @property
    def limits(self):
        """list: List of (lower, upper) bounds for the bounding box"""
        theta = self.angle_rad
        tan_t = np.tan(theta)

        tan_tx = - self.b / self.a * tan_t
        tx_max = np.arctan(tan_tx)
        tx_min = np.pi + tx_max

        if np.isclose(tan_t, 0) and np.cos(theta) > 0:
            ty_max = 0.5 * np.pi
            ty_min = - 0.5 * np.pi
        elif np.isclose(tan_t, 0) and np.cos(theta) < 0:
            ty_max = -0.5 * np.pi
            ty_min = 0.5 * np.pi
        else:
            tan_ty = self.b / (self.a * np.tan(theta))
            ty_max = (np.arctan(tan_ty) + np.pi) % np.pi
            ty_min = np.pi + ty_max

        def xp(t):
            return self.a * np.cos(t)

        def yp(t):
            return self.b * np.sin(t)

        def xr(t):
            return np.cos(theta) * xp(t) - np.sin(theta) * yp(t)

        def yr(t):
            return np.sin(theta) * xp(t) + np.cos(theta) * yp(t)

        x_max = xr(tx_max) + self.center[0]
        x_min = xr(tx_min) + self.center[0]

        y_max = yr(ty_max) + self.center[1]
        y_min = yr(ty_min) + self.center[1]

        return [sorted((x_min, x_max)), sorted((y_min, y_max))]

    @property
    def sample_limits(self):
        """list: List of (lower, upper) bounds for the sampling region"""
        return self.limits

    # ----------------------------------------------------------------------- #
    # Within Test                                                             #
    # ----------------------------------------------------------------------- #
    def within(self, points):
        """Test if points are within ellipse.

        This function tests whether a point or set of points are within the
        ellipse. For the set of points, a list of booleans is returned to
        indicate which points are within the ellipse.

        Args:
            points (list or numpy.ndarray): Point or list of points.

        Returns:
            bool or numpy.ndarray: Set to True for points in ellipse.

        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        rot_pos = rel_pos.dot(self.orientation)
        scl_pos = rot_pos / np.array(self.axes).reshape(1, -1)

        sq_dist = np.sum(scl_pos * scl_pos, axis=-1)

        mask = sq_dist <= 1
        if single_pt:
            return mask[0]
        else:
            return mask

    # ----------------------------------------------------------------------- #
    # Reflect                                                                 #
    # ----------------------------------------------------------------------- #
    def reflect(self, points):
        """Reflect points across surface.

        This function reflects a point or set of points across the surface
        of the ellipse. Points at the center of the ellipse are not
        reflected.

        Args:
            points (list or numpy.ndarray): Nx2 list of points to reflect.

        Returns:
            numpy.ndarray: Reflected points.

        """
        pts = np.array(points)
        single_pt = pts.ndim == 1
        if single_pt:
            pts = pts.reshape(1, -1)

        rel_pos = pts - np.array(self.center)
        rot_pos = rel_pos.dot(self.orientation)
        scl_pos = rot_pos / np.array(self.axes).reshape(1, -1)

        dist = np.sqrt(np.sum(scl_pos * scl_pos, axis=-1))
        mask = dist > 0
        if not np.any(mask):
            return np.array([])

        new_dist = 2 - dist[mask]
        scl = new_dist / dist[mask]

        new_scl_pos = scl_pos[mask] * scl
        new_rel_pos = new_scl_pos.dot(self.orientation.T)
        new_pos = new_rel_pos + np.array(self.center)

        if single_pt:
            return new_pos[0]
        else:
            return new_pos



# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
from __future__ import division

import numpy as np
from matplotlib import patches
from matplotlib import pyplot as plt

from microstructpy.geometry.n_sphere import NSphere

__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# Circle Class                                                                #
#                                                                             #
# --------------------------------------------------------------------------- #
class Circle(NSphere):
    """A 2D circle.

    This class represents a two-dimensional circle. It is defined by
    a center point and size parameter, which can be either radius or diameter.

    Without parameters, this returns a unit circle centered on the origin.

    Args:
        r (float): *(optional)* The radius of the circle. Defaults to 1.
        center (list): *(optional)* The coordinates of the center.
            Defaults to (0, 0).
        diameter : *(optional)* Alias for ``2*r``.
        radius : *(optional)* Alias for ``r``.
        d : *(optional)* Alias for ``2*r``.
        size : *(optional)* Alias for ``2*r``.
        position : *(optional)* Alias for ``center``.
    """
    # ----------------------------------------------------------------------- #
    # Constructor                                                             #
    # ----------------------------------------------------------------------- #
    def __init__(self, **kwargs):
        if 'area' in kwargs:
            a = kwargs['area']
            r = np.sqrt(a / np.pi)
            kwargs['r'] = r

        NSphere.__init__(self, **kwargs)
        if len(self.center) == 0:
            self.center = tuple(self.n_dim * [0])

    # ----------------------------------------------------------------------- #
    # Representation Function                                                 #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'Circle('
        repr_str += 'center=' + repr(tuple(self.center))
        repr_str += ', radius=' + repr(self.r)
        repr_str += ')'
        return repr_str

    # ----------------------------------------------------------------------- #
    # Number of Dimensions                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def n_dim(self):
        """int: number of dimensions, 2"""
        return 2

    # ----------------------------------------------------------------------- #
    # Area                                                                    #
    # ----------------------------------------------------------------------- #
    @property
    def area(self):
        r"""float: area of cirle, :math:`A=\pi r^2`"""
        return np.pi * self.r * self.r

    @property
    def volume(self):
        """float: alias for area"""
        return self.area

    # ----------------------------------------------------------------------- #
    # Expected Area                                                           #
    # ----------------------------------------------------------------------- #
    @classmethod
    def area_expectation(cls, **kwargs):
        r"""Expected value of area.

        This function computes the expected value for the area of a circle.
        The keyword arguments are the same as the class parameters.
        The values can be constants (ints or floats), or a distribution from
        the SciPy :mod:`scipy.stats` module.

        The expected value is computed by the following formula:

        .. math::

            \mathbb{E}[A] = \pi \mathbb{E}[R^2] = \pi (\mu_R^2 + \sigma_R^2)


        For example::

            >>> from microstructpy.geometry import Circle
            >>> Circle.area_expectation(r=1)
            3.141592653589793
            >>> from scipy.stats import norm
            >>> Circle.area_expectation(r=norm(1, 1))
            6.283185307179586

        Args:
            **kwargs: Keyword arguments, see :class:`.Circle`.

        Returns:
            float: Expected value of the area of the circle.

        """  # NOQA: E501
        # Check for radius distribution
        r_dist = None
        if 'radius' in kwargs:
            r_dist = kwargs['radius']
        elif 'r' in kwargs:
            r_dist = kwargs['r']

        try:
            return np.pi * r_dist.moment(2)
        except AttributeError:
            try:
                return np.pi * r_dist * r_dist
            except TypeError:
                pass

        # Check for diameter distribution
        d_dist = None
        for d_kw in ('d', 'diameter', 'size'):
            if d_kw in kwargs:
                d_dist = kwargs[d_kw]
                break

        try:
            return 0.25 * np.pi * d_dist.moment(2)
        except AttributeError:
            try:
                return 0.25 * np.pi * d_dist * d_dist
            except TypeError:
                pass

        if 'area' in kwargs:
            a_dist = kwargs['area']
            try:
                a_exp = a_dist.moment(1)
            except AttributeError:
                a_exp = a_dist
            return a_exp

        # Raise error
        e_str = 'Could not find one of the following keywords in the inputs: '
        e_str += 'r, radius, d, diameter, size.'
        raise KeyError(e_str)

    # ----------------------------------------------------------------------- #
    # Plot Function                                                           #
    # ----------------------------------------------------------------------- #
    def plot(self, **kwargs):
        """Plot the circle.

        This function adds a :class:`matplotlib.patches.Circle` to the
        current axes. The keyword arguments are passed through to the
        circle patch.

        Args:
            **kwargs (dict): Keyword arguments for matplotlib.

        """  # NOQA: E501
        c = patches.Circle(self.center, self.r, **kwargs)
        plt.gca().add_patch(c)



from __future__ import division

import os

import numpy as np
from matplotlib import collections
from matplotlib import pyplot as plt
from matplotlib.offsetbox import AnnotationBbox
from matplotlib.offsetbox import OffsetImage

import microstructpy as msp


def main(n_seeds, size_rng, pos_rng, k_lw):
    bkgrnd_color = 'black'
    line_color = (1, 1, 1, 1)  # white

    dpi = 300
    init_size = 2000
    logo_size = 1500
    favicon_size = 48

    logo_basename = 'logo.svg'
    favicon_basename = 'favicon.ico'
    social_basename = 'social.png'
    file_dir = os.path.dirname(os.path.realpath(__file__))
    path = os.path.join(file_dir, 'logo')
    if not os.path.exists(path):
        os.makedirs(path)
    print(path)
    logo_filename = os.path.join(path, logo_basename)
    pad_filename = os.path.join(path, 'pad_' + logo_basename)
    favicon_filename = os.path.join(path, favicon_basename)
    social_filename = os.path.join(path, social_basename)

    # Set Domain
    domain = msp.geometry.Circle()

    # Set Seed List
    np.random.seed(size_rng)
    rs = 0.3 * np.random.rand(n_seeds)

    factory = msp.seeding.Seed.factory
    seeds = msp.seeding.SeedList([factory('circle', r=r) for r in rs])
    seeds.position(domain, rng_seed=pos_rng)

    # Create the Poly Mesh
    pmesh = msp.meshing.PolyMesh.from_seeds(seeds, domain)

    # Create and Format the Figure
    plt.clf()
    plt.close('all')
    fig = plt.figure(figsize=(init_size / dpi, init_size / dpi), dpi=dpi)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    fig.add_axes(ax)

    # Plot the Domain
    domain.plot(ec='none', fc=bkgrnd_color)

    # Plot the Facets
    facet_colors = []
    for neigh_pair in pmesh.facet_neighbors:
        if min(neigh_pair) < 0:
            facet_colors.append('none')
        else:
            facet_colors.append(line_color)

    lw = k_lw * init_size / 100
    pmesh.plot_facets(index_by='facet', colors=facet_colors,
                      linewidth=lw, capstyle='round')

    pts = np.array(pmesh.points)
    rs = np.sqrt(np.sum(pts * pts, axis=1))
    mask = np.isclose(rs, 1)

    edges = []
    for facet in pmesh.facets:
        if np.sum(mask[facet]) != 1:
            continue

        edge = np.copy(pts[facet])
        if mask[facet[0]]:
            u = edge[0] - edge[1]
            u *= 1.1
            edge[0] = edge[1] + u
        else:
            u = edge[1] - edge[0]
            u *= 1.1
            edge[1] = edge[0] + u
        edges.append(edge)

    pc = collections.LineCollection(edges, color=line_color, linewidth=lw,
                                    capstyle='round')
    ax.add_collection(pc)

    # Format the Plot and Convert to Image Array
    plt.axis('square')
    plt.axis(1.01 * np.array([-1, 1, -1, 1]))
    fig.canvas.draw()

    plt_im = np.array(fig.canvas.get_renderer()._renderer)
    mask = plt_im[:, :, 0] > 0.5 * 255

    # Create the Logo
    logo_im = np.copy(plt_im)

    xx, yy = np.meshgrid(*[np.arange(n) for n in logo_im.shape[:2]])
    zz = - 0.2 * xx + 0.9 * yy
    ss = (zz - zz.min()) / (zz.max() - zz.min())

    c1 = [67, 206, 162]
    c2 = [24, 90, 157]

    logo_im[mask, -1] = 0  # transparent background

    # gradient
    for i in range(logo_im.shape[-1] - 1):
        logo_im[~mask, i] = (1 - ss[~mask]) * c1[i] + ss[~mask] * c2[i]

    inds = np.linspace(0, logo_im.shape[0] - 1, logo_size).astype('int')
    logo_im = logo_im[inds]
    logo_im = logo_im[:, inds]

    pad_w = logo_im.shape[0]
    pad_h = 0.5 * logo_im.shape[1]
    pad_shape = np.array([pad_w, pad_h, logo_im.shape[2]]).astype('int')
    logo_pad = np.zeros(pad_shape, dtype=logo_im.dtype)
    pad_im = np.concatenate((logo_pad, logo_im, logo_pad), axis=1)
    doc_im = np.concatenate((logo_pad, pad_im, logo_pad), axis=1)

    plt.imsave(logo_filename, logo_im, dpi=dpi)
    plt.imsave(logo_filename.replace('.svg', '.png'), logo_im, dpi=dpi)
    plt.imsave(pad_filename, pad_im, dpi=dpi)
    plt.imsave(pad_filename.replace('.svg', '.png'), doc_im, dpi=dpi)

    # Create the Favicon
    fav_im = np.copy(logo_im)
    inds = np.linspace(0, fav_im.shape[0] - 1, favicon_size).astype('int')
    fav_im = fav_im[inds]
    fav_im = fav_im[:, inds]

    plt.imsave(favicon_filename, fav_im, dpi=dpi, format='png')

    # Create the Social Banner
    fig_social, ax_social = plt.subplots()

    ax_social.set_xlim(0, 2)
    ax_social.set_ylim(0, 1)
    ax_social.set_aspect('equal')

    ax_social.set_axis_off()
    ax_social.get_xaxis().set_visible(False)
    ax_social.get_yaxis().set_visible(False)

    imagebox = OffsetImage(logo_im, zoom=0.05)
    ab = AnnotationBbox(imagebox, (1, 0.7), frameon=False)
    ax_social.add_artist(ab)
    ax_social.text(1, 0.35, 'MicroStructPy',
                   fontsize=20,
                   weight='bold',
                   horizontalalignment='center',
                   verticalalignment='center')
    ax_social.text(1, 0.23, 'Microstructure Mesh Generation in Python',
                   fontsize=10,
                   horizontalalignment='center',
                   verticalalignment='center')
    plt.draw()
    plt.savefig(social_filename, bbox_inches='tight')
    plt.close('all')


if __name__ == '__main__':
    n_seeds = 14
    size_rng = 4
    pos_rng = 7
    k_lw = 1.1

    main(n_seeds, size_rng, pos_rng, k_lw)



import numpy as np

from microstructpy import _misc


def test_from_str_int():
    pairs = [('0', 0),
             ('1', 1),
             ('2', 2),
             ('-1', -1),
             ('-2', -2)]

    for int_str, int_exp in pairs:
        int_act = _misc.from_str(int_str)
        a_str = 'Expected ' + str(int_exp) + ' and got ' + str(int_act)
        assert int_exp == int_act, a_str


def test_from_str_float():
    pairs = [('1.234', 1.234),
             ('0.214', 0.214),
             ('-0.4', -0.4),
             ('1.2e5', 1.2e5)]

    for flt_str, flt_exp in pairs:
        flt_act = _misc.from_str(flt_str)
        a_str = 'Expected ' + str(flt_exp) + ' and got ' + str(flt_act)
        assert flt_exp == flt_act, a_str


def test_from_str_bool():
    pairs = [('True', True),
             ('False', False),
             ('true', True),
             ('false', False)]

    for bool_str, bool_exp in pairs:
        bool_act = _misc.from_str(bool_str)
        a_str = 'Expected ' + str(bool_exp) + ' and got ' + str(bool_act)
        a_str += ' for string ' + repr(bool_str)
        assert bool_exp == bool_act, a_str


def test_from_str_list():
    pairs = [('[0]', [0]),
             ('[1, 0, a]', [1, 0, 'a']),
             ('-2.3, true', [-2.3, True])]

    for list_str, list_exp in pairs:
        list_act = _misc.from_str(list_str)
        assert len(list_exp) == len(list_act)
        for act_val, exp_val in zip(list_act, list_exp):
            assert act_val == exp_val


def test_from_str_list_of_lists():
    lol_str = '[[1, 0, 0, True, False], [2, 4, a, -2.3]]'
    lol_exp = [[1, 0, 0, True, False], [2, 4, 'a', -2.3]]

    lol_act = _misc.from_str(lol_str)

    assert len(lol_exp) == len(lol_act)
    for list_exp, list_act in zip(lol_exp, lol_act):
        assert len(list_exp) == len(list_act)
        for val_exp, val_act in zip(list_exp, list_act):
            assert val_exp == val_act


def test_tangent_sphere_2D():
    pts = np.array([(0, 0), (4, 0), (3, 3)])
    rads = np.array([1, 0.5, 0.5])

    simps = [None, np.array([[0, 1, 2]])]
    for simp in simps:
        x, y, rad = _misc.tangent_sphere(pts, rads, simplices=simp)
        cen = np.array([x, y])

        rel_pos = pts - cen
        dist = np.sqrt(np.sum(rel_pos * rel_pos, axis=-1))
        assert np.all(np.isclose(dist, rads + rad))


def test_tangent_sphere_ND():
    n = 7
    np.random.seed(0)
    pts = np.random.rand(n + 1, n)
    nsphere = _misc.tangent_sphere(pts)
    cen = nsphere[:-1]
    rad = nsphere[-1]

    rel_pos = pts - cen
    dist = np.sqrt(np.sum(rel_pos * rel_pos, axis=-1))
    assert np.all(np.isclose(dist, rad))



import os

from matplotlib import pyplot as plt

import microstructpy as msp

# Create domain
domain = msp.geometry.Square()

# Create list of seed points
factory = msp.seeding.Seed.factory
n = 50
seeds = msp.seeding.SeedList([factory('circle', r=0.01) for i in range(n)])
seeds.position(domain)

# Create Voronoi diagram
pmesh = msp.meshing.PolyMesh.from_seeds(seeds, domain)

# Plot Voronoi diagram and seed points
pmesh.plot(edgecolors='k', facecolors='gray')
seeds.plot(edgecolors='k', facecolors='none')

plt.axis('square')
plt.xlim(domain.limits[0])
plt.ylim(domain.limits[1])

file_dir = os.path.dirname(os.path.realpath(__file__))
filename = os.path.join(file_dir, 'standard_voronoi/voronoi_diagram.png')
dirs = os.path.dirname(filename)
if not os.path.exists(dirs):
    os.makedirs(dirs)
plt.savefig(filename, bbox_inches='tight', pad_inches=0)



from __future__ import division

import os

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt
from scipy.spatial import distance

import microstructpy as msp

# Create domain
domain = msp.geometry.Square(corner=(0, 0))

# Create list of seed points
factory = msp.seeding.Seed.factory
n = 200
seeds = msp.seeding.SeedList([factory('circle', r=0.007) for i in range(n)])

# Position seeds according to Mitchell's Best Candidate Algorithm
np.random.seed(0)

lims = np.array(domain.limits) * (1 - 1e-5)
centers = np.zeros((n, 2))

for i in range(n):
    f = np.random.rand(i + 1, 2)
    pts = f * lims[:, 0] + (1 - f) * lims[:, 1]
    try:
        min_dists = distance.cdist(pts, centers[:i]).min(axis=1)
        i_max = np.argmax(min_dists)
    except ValueError:  # this is the case when i=0
        i_max = 0
    centers[i] = pts[i_max]
    seeds[i].position = centers[i]

# Create Voronoi diagram
pmesh = msp.meshing.PolyMesh.from_seeds(seeds, domain)

# Set colors based on area
areas = pmesh.volumes
std_area = domain.area / n
min_area = min(areas)
max_area = max(areas)
cell_colors = np.zeros((n, 3))
for i in range(n):
    if areas[i] < std_area:
        f_red = (areas[i] - min_area) / (std_area - min_area)
        f_green = (areas[i] - min_area) / (std_area - min_area)
        f_blue = 1
    else:
        f_red = 1
        f_green = (max_area - areas[i]) / (max_area - std_area)
        f_blue = (max_area - areas[i]) / (max_area - std_area)
    cell_colors[i] = (f_red, f_green, f_blue)

# Create colorbar
vs = (std_area - min_area) / (max_area - min_area)
colors = [(0, (0, 0, 1)), (vs, (1, 1, 1)), (1, (1, 0, 0))]
cmap = mpl.colors.LinearSegmentedColormap.from_list('area_cmap', colors)
norm = mpl.colors.Normalize(vmin=min_area, vmax=max_area)
sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
sm.set_array([])
cb = plt.colorbar(sm, ticks=[min_area, std_area, max_area],
                  orientation='horizontal', fraction=0.046, pad=0.08)
cb.set_label('Cell Area')

# Plot Voronoi diagram and seed points
pmesh.plot(edgecolors='k', facecolors=cell_colors)
seeds.plot(edgecolors='k', facecolors='none')

plt.axis('square')
plt.xlim(domain.limits[0])
plt.ylim(domain.limits[1])

# Save diagram
file_dir = os.path.dirname(os.path.realpath(__file__))
filename = os.path.join(file_dir, 'uniform_seeding/voronoi_diagram.png')
dirs = os.path.dirname(filename)
if not os.path.exists(dirs):
    os.makedirs(dirs)
plt.savefig(filename, bbox_inches='tight', pad_inches=0)



import os
import shutil

import numpy as np
from matplotlib import image as mpim
from matplotlib import pyplot as plt

import microstructpy as msp

# Read in image
image_basename = 'aluminum_micro.png'
image_path = os.path.dirname(__file__)
image_filename = os.path.join(image_path, image_basename)
image = mpim.imread(image_filename)
im_brightness = image[:, :, 0]

# Bin the pixels
br_bins = [0.00, 0.50, 1.00]

bin_nums = np.zeros_like(im_brightness, dtype='int')
for i in range(len(br_bins) - 1):
    lb = br_bins[i]
    ub = br_bins[i + 1]
    mask = np.logical_and(im_brightness >= lb, im_brightness <= ub)
    bin_nums[mask] = i

# Define the phases
phases = [{'color': c, 'material_type': 'amorphous'} for c in ('C0', 'C1')]

# Create the polygon mesh
m, n = bin_nums.shape
x = np.arange(n + 1).astype('float')
y = m + 1 - np.arange(m + 1).astype('float')
xx, yy = np.meshgrid(x, y)
pts = np.array([xx.flatten(), yy.flatten()]).T
kps = np.arange(len(pts)).reshape(xx.shape)

n_facets = 2 * (m + m * n + n)
n_regions = m * n
facets = np.full((n_facets, 2), -1)
regions = np.full((n_regions, 4), 0)
region_phases = np.full(n_regions, 0)

facet_top = np.full((m, n), -1, dtype='int')
facet_bottom = np.full((m, n), -1, dtype='int')
facet_left = np.full((m, n), -1, dtype='int')
facet_right = np.full((m, n), -1, dtype='int')

k_facets = 0
k_regions = 0
for i in range(m):
    for j in range(n):
        kp_top_left = kps[i, j]
        kp_bottom_left = kps[i + 1, j]
        kp_top_right = kps[i, j + 1]
        kp_bottom_right = kps[i + 1, j + 1]

        # left facet
        if facet_left[i, j] < 0:
            fnum_left = k_facets
            facets[fnum_left] = (kp_top_left, kp_bottom_left)
            k_facets += 1

            if j > 0:
                facet_right[i, j - 1] = fnum_left
        else:
            fnum_left = facet_left[i, j]

        # right facet
        if facet_right[i, j] < 0:
            fnum_right = k_facets
            facets[fnum_right] = (kp_top_right, kp_bottom_right)
            k_facets += 1

            if j + 1 < n:
                facet_left[i, j + 1] = fnum_right
        else:
            fnum_right = facet_right[i, j]

        # top facet
        if facet_top[i, j] < 0:
            fnum_top = k_facets
            facets[fnum_top] = (kp_top_left, kp_top_right)
            k_facets += 1

            if i > 0:
                facet_bottom[i - 1, j] = fnum_top
        else:
            fnum_top = facet_top[i, j]

        # bottom facet
        if facet_bottom[i, j] < 0:
            fnum_bottom = k_facets
            facets[fnum_bottom] = (kp_bottom_left, kp_bottom_right)
            k_facets += 1

            if i + 1 < m:
                facet_top[i + 1, j] = fnum_bottom
        else:
            fnum_bottom = facet_bottom[i, j]

        # region
        region = (fnum_top, fnum_left, fnum_bottom, fnum_right)
        regions[k_regions] = region
        region_phases[k_regions] = bin_nums[i, j]
        k_regions += 1


pmesh = msp.meshing.PolyMesh(pts, facets, regions,
                             seed_numbers=range(n_regions),
                             phase_numbers=region_phases)

# Create the triangle mesh
tmesh = msp.meshing.TriMesh.from_polymesh(pmesh, phases=phases, min_angle=20)

# Plot triangle mesh
fig = plt.figure()
ax = plt.Axes(fig, [0., 0., 1., 1.])
ax.set_axis_off()
ax.get_xaxis().set_visible(False)
ax.get_yaxis().set_visible(False)
fig.add_axes(ax)

fcs = [phases[region_phases[r]]['color'] for r in tmesh.element_attributes]
tmesh.plot(facecolors=fcs, edgecolors='k', lw=0.2)


plt.axis('square')
plt.xlim(x.min(), x.max())
plt.ylim(y.min(), y.max())
plt.axis('off')

# Save plot and copy input file
plot_basename = 'from_image/trimesh.png'
file_dir = os.path.dirname(os.path.realpath(__file__))
filename = os.path.join(file_dir, plot_basename)
dirs = os.path.dirname(filename)
if not os.path.exists(dirs):
    os.makedirs(dirs)
plt.savefig(filename, bbox_inches='tight', pad_inches=0)

shutil.copy(image_filename, dirs)



# -*- coding: utf-8 -*-
r"""
Plot all demos

"""

import glob
import locale
import os
import shutil
import subprocess

import matplotlib.pyplot as plt
import microstructpy as msp

locale.setlocale(locale.LC_NUMERIC, "C")

example_dir = '../../../src/microstructpy/examples'

welcome_fnames = ['intro_2_quality/trimesh.png',
                  'minimal/polymesh.png',
                  'basalt_circle/trimesh.png',
                  'foam/trimesh.png',
                  'two_phase_3D/trimesh.png',
                  'colormap/trimesh.png']

def main():
    # Copy Supporting Files
    supporting_files = ['aphanitic_cdf.csv', 'olivine_cdf.csv']
    for fname in supporting_files:
        filename = os.path.join(example_dir, fname)
        shutil.copy(filename, '.')

    # Run XML files
    xml_pattern = os.path.join(example_dir, '*.xml')
    for filename in glob.glob(xml_pattern):
        msp.cli.run_file(filename)

    # Run Python Scripts
    py_pattern = os.path.join(example_dir, '*.py')
    for filename in glob.glob(py_pattern):
        subprocess.call(['python', filename])

    # Remove Supporting Files
    for fname in supporting_files:
        os.remove(fname)

    # Create welcome figure
    create_welcome()

    # Create example subfigures
    seed_poly_tri('.')


def create_welcome():
    fig, axes = plt.subplots(2, 3, figsize=(21, 15))
    plt.subplots_adjust(wspace=0.05, hspace=0)
    for i, fname in enumerate(welcome_fnames):
        filename = os.path.join(example_dir, fname)
        im = plt.imread(filename)
        
        row_num = int(i / 3)
        col_num = i % 3
        ax = axes[row_num, col_num]
        ax.imshow(im)

        ax.set_axis_off()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    sub_fname = 'welcome_examples.png'
    sub_filename = os.path.join(example_dir, sub_fname)
    plt.savefig(sub_filename, pad_inches=0, bbox_inches='tight', dpi=200)
    plt.close('all')


def seed_poly_tri(filepath):
    basenames = ['seeds.png', 'polymesh.png', 'trimesh.png']
    ex_path = os.path.join(example_dir, filepath)
    fig, axes = plt.subplots(1, 3, figsize=(20, 10))
    plt.subplots_adjust(wspace=0.05, hspace=0)
    for i, fname in enumerate(basenames):
        filename = os.path.join(ex_path, fname)
        im = plt.imread(filename)

        ax = axes[i]
        ax.imshow(im)

        ax.set_axis_off()
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
    
    sub_fname = 'joined.png'
    sub_filename = os.path.join(ex_path, sub_fname)
    plt.savefig(sub_filename, pad_inches=0, bbox_inches='tight', dpi=300)
    plt.close('all')


if __name__ == '__main__':
    main()
    





"""
Plot figures relevant to :class:`microstructpy.geometry.Ellipse`

"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import collections
from matplotlib.ticker import FormatStrFormatter

import microstructpy as msp


def main():
    breakdown()  # ellipse_001.png - ellipse breakdown


def breakdown():
    a = 3
    b = 1
    x1 = 0.7

    plt.figure(figsize=(14, 6))
    ellipse = msp.geometry.Ellipse(a=a, b=b)
    approx = ellipse.approximate(x1)
    ellipse.plot(edgecolor='k', facecolor='none', lw=3)
    t = np.linspace(0, 2 * np.pi)
    for x, y, r in approx:
        plt.plot(x + r * np.cos(t), y + r * np.sin(t), 'b')
    
    xticks = np.unique(np.concatenate((approx[:, 0], (-a, a))))
    plt.xticks(xticks)
    plt.yticks(np.unique(np.concatenate((approx[:, 1], (-b, b)))))
    plt.gca().set_xticklabels([str(round(float(label), 1)) for label in xticks])
    plt.axis('scaled')
    plt.grid(True, linestyle=':')


if __name__ == '__main__':
    plt.rc('savefig', dpi=300, bbox='tight', pad_inches=0)
    main()


"""Polygon Meshing

This module contains the class definition for the PolyMesh class.

"""
# --------------------------------------------------------------------------- #
#                                                                             #
# Import Modules                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #


from __future__ import division
from __future__ import print_function

import os
import subprocess
import sys
import tempfile
import warnings

import numpy as np
import pyvoro
from matplotlib import collections
from matplotlib import patches
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy.spatial import distance

from microstructpy import _misc
from microstructpy import geometry

__all__ = ['PolyMesh']
__author__ = 'Kenneth (Kip) Hart'


# --------------------------------------------------------------------------- #
#                                                                             #
# PolyMesh Class                                                              #
#                                                                             #
# --------------------------------------------------------------------------- #
class PolyMesh(object):
    """Polygonal/Polyhedral mesh.

    The PolyMesh class contains the points, edges, regions, etc. in a polygon
    (2D) or polyhedron (3D) mesh.

    The points attribute is a numpy array containing the (x, y) or (x, y, z)
    coordinates of each point in the mesh. This is the only attribute that
    contains floating point numbers. The rest contain indices/integers.

    The facets attribute describes the interfaces between the polygons/
    polyhedra. In 2D, these interfaces are line segments and each facet
    contains the indices of the points at each end of the line segment. These
    indices are unorderd. In 3D, the interfaces are polygons so each facet
    contains the indices of the points on that polygon. These indices are
    ordered such that neighboring keypoints are connected by line segments
    that form the polygon.

    The regions attribute contains the area (2D) or volume (3D). In 2D, a
    region is given by an ordered list of facets, or edges, that enclose the
    polygon. In 3D, the region is given by an un-ordered list of facets,
    or polygons, that enclose the polyhedron.

    For each region, there is also an associated seed number and material
    phase. These data are stored in the seed_number and phase_number
    attributes, which have the same length as the regions list.

    Args:
        points (list or numpy.ndarray): An Nx2 or Nx3 array of coordinates
            in the mesh.
        facets (list): List of facets between regions. In 2D, this is a list
            of edges (Nx2). In 3D, this is a list of 3D polygons.
        regions (list): A list of polygons (2D) or polyhedra (3D), with each
            element of the list being a list of facet indices.
        seed_numbers (list or numpy.ndarray): *(optional)* The seed number
            associated with each region.
            Defaults to 0 for all regions.
        phase_numbers (list or numpy.ndarray): *(optional)* The phase number
            associated with each region.
            Defaults to 0 for all regions.
        facet_neighbors (list or numpy.ndarray): *(optional)* The region
            numbers on either side of each facet.
            If not givien, a neighbor list is computed from ``regions``.
        volumes (list or numpy.ndarray): *(optional)* The area/volume of each
            region.
            If not given, region volumes are calculated based on ``points``,
            ``facets``, and ``regions``.

    """

    # ----------------------------------------------------------------------- #
    # Constructors                                                            #
    # ----------------------------------------------------------------------- #
    def __init__(self, points, facets, regions, seed_numbers=None,
                 phase_numbers=None, facet_neighbors=None, volumes=None):

        self.points = points
        self.facets = facets
        self.regions = regions

        if facet_neighbors is None:
            # Find facet neighbors
            facet_neighs = [[-1, -1] for _ in facets]
            n_neighs = [0 for _ in facets]
            for r_num, region in enumerate(regions):
                for f_num in region:
                    ind = n_neighs[f_num]
                    facet_neighs[f_num][ind] = r_num
                    n_neighs[f_num] += 1

            # update negative neighbor numbers to follow the voro++
            # convention, described in the %n section of this website:
            # http://math.lbl.gov/voro++/doc/custom.html
            pt_arr = np.array(points)
            pt_mins = pt_arr.min(axis=0)
            pt_maxs = pt_arr.max(axis=0)
            for fnum, facet in enumerate(facets):
                if facet_neighs[fnum][-1] != -1:
                    continue

                f_pts = pt_arr[facet, :]
                min_match = np.all(np.isclose(f_pts, pt_mins), axis=0)
                if np.any(min_match):
                    ld = 3 - len(min_match)
                    mask = np.pad(min_match, (0, ld), 'constant',
                                  constant_values=(False, False))
                    id = np.array([-1, -3, -5])[mask][0]
                    facet_neighs[fnum][-1] = id

                max_match = np.all(np.isclose(f_pts, pt_maxs), axis=0)
                if np.any(max_match):
                    ld = 3 - len(max_match)
                    mask = np.pad(max_match, (0, ld), 'constant',
                                  constant_values=(False, False))
                    id = np.array([-2, -4, -6])[mask][0]
                    facet_neighs[fnum][-1] = id

            self.facet_neighbors = facet_neighs
        else:
            self.facet_neighbors = facet_neighbors

        if seed_numbers is None:
            self.seed_numbers = [0 for _ in regions]
        else:
            self.seed_numbers = seed_numbers

        if phase_numbers is None:
            self.phase_numbers = [0 for _ in regions]
        else:
            self.phase_numbers = phase_numbers

        if volumes is None:
            vols = np.zeros(len(self.regions))
            n = len(self.points[0])
            for i, region in enumerate(self.regions):
                # 'center' of region is arbitrary, since region is convex
                cen = np.array(self.points)[self.facets[region[0]][0]]
                for f_num in region:
                    facet = np.array(self.facets[f_num])
                    j_max = len(facet) - n + 2
                    # convert facet into (n-1)D simplices
                    for j in range(1, j_max):
                        inds = np.append(np.arange(j, j + n - 1), 0)
                        simplex = facet[inds]
                        facet_pts = np.array(self.points)[simplex]
                        rel_pos = facet_pts - cen
                        # simplex volume is |det([Dx1, Dy1; Dx2, Dy2])| in 2D
                        vols[i] += np.abs(np.linalg.det(rel_pos))

            # the 1/2 out front in 2D and 1/6 in 3D
            while n > 1:
                vols /= n
                n -= 1
            self.volumes = vols
        else:
            self.volumes = volumes

    # ----------------------------------------------------------------------- #
    # Representation and String Functions                                     #
    # ----------------------------------------------------------------------- #
    def __repr__(self):
        repr_str = 'PolyMesh('
        repr_str += repr(self.points)
        repr_str += ', '
        repr_str += repr(self.facets)
        repr_str += ', '
        repr_str += repr(self.regions)
        for att in ('seed_numbers', 'phase_numbers'):
            repr_str += ', '
            vals = self.__dict__[att]
            if all([n == 0 for n in vals]):
                repr_str += repr(None)
            else:
                repr_str += repr(vals)
        repr_str += ')'
        return repr_str

    def __str__(self):
        nv = len(self.points)
        nd = len(self.points[0])
        pt_fmt = '\t'
        pt_fmt += ', '.join(['{pt[' + str(i) + ']: e}' for i in range(nd)])

        str_str = 'Mesh Points: ' + str(nv) + '\n'
        str_str += ''.join([pt_fmt.format(pt=p) + '\n' for p in self.points])

        str_str += 'Mesh Facets: ' + str(len(self.facets)) + '\n'
        str_str += ''.join(['\t' + str(tuple(f))[1:-1] + '\n'
                            for f in self.facets])

        str_str += 'Facet Neighbors: ' + str(len(self.facet_neighbors)) + '\n'
        str_str += ''.join(['\t' + str(tuple(n))[1:-1] + '\n'
                            for n in self.facet_neighbors])

        str_str += 'Mesh Regions: ' + str(len(self.regions)) + '\n'
        str_str += ''.join(['\t' + str(tuple(r))[1:-1] + '\n'
                            for r in self.regions])

        str_str += 'Seed Numbers: ' + str(len(self.seed_numbers)) + '\n'
        str_str += ''.join(['\t' + str(n) + '\n' for n in self.seed_numbers])

        str_str += 'Phase Numbers: ' + str(len(self.phase_numbers)) + '\n'
        str_str += ''.join(['\t' + str(n) + '\n' for n in self.phase_numbers])

        str_str += 'Volumes: ' + str(len(self.volumes)) + '\n'
        str_str += '\n'.join(['\t' + str(v) for v in self.volumes])
        return str_str

    # ----------------------------------------------------------------------- #
    # Read and Write Functions                                                #
    # ----------------------------------------------------------------------- #
    def write(self, filename, format='txt'):
        """Write the mesh to a file.

        This function writes the polygon/polyhedron mesh to a file.
        See the :ref:`s_poly_file_io` section of the
        :ref:`c_file_formats` guide for more information about the available
        output file formats.

        Args:
            filename (str): Name of the file to be written.
            format (str): *(optional)* {'txt' | 'poly' | 'ply' | 'vtk' }
                Format of the data in the file. Defaults to ``'txt'``.

        """
        if format in ('str', 'txt'):
            with open(filename, 'w') as f:
                f.write(str(self) + '\n')

        elif format == 'poly':
            nv = len(self.points)
            nd = len(self.points[0])
            nf = len(self.facets)
            assert nd == 2

            poly = '# Polygon Mesh\n'
            poly += ' '.join([str(n) for n in (nv, 2, 0, 0)]) + '\n'

            # vertices
            poly += '# Vertices\n'
            poly += ''.join([str(i) + ''.join([' {: e}'.format(x) for x in pt])
                             + '\n' for i, pt in enumerate(self.points)])

            # facets
            poly += '# Segments\n'
            poly += ' '.join([str(n) for n in (nf, 0)]) + '\n'
            poly += ''.join([' '.join([str(n) for n in (nv + i, k1, k2)])
                             + '\n' for i, (k1, k2) in enumerate(self.facets)])

        elif format == 'ply':
            nv = len(self.points)
            nd = len(self.points[0])
            nf = len(self.facets)
            assert nd <= 3
            axes = ['x', 'y', 'z'][:nd]

            # header
            ply = 'ply\n'
            ply += 'format ascii 1.0\n'
            ply += 'element vertex ' + str(nv) + '\n'
            ply += ''.join(['property float32 ' + a + '\n' for a in axes])
            ply += 'element face ' + str(nf) + '\n'
            ply += 'property list uint8 int32 vertex_indices\n'
            ply += 'end_header\n'

            # vertices
            ply += ''.join([' '.join(['{: e}'.format(x) for x in pt]) + '\n'
                            for pt in self.points])

            # faces
            ply += ''.join([str(len(f)) + ''.join([' ' + str(kp) for kp in f])
                            + '\n' for f in self.facets])

            with open(filename, 'w') as f:
                f.write(ply)

        elif format == 'vtk':
            vtk_s = '# vtk DataFile Version 2.0\n'
            vtk_s += 'Polygonal Mesh\n'
            vtk_s += 'ASCII\n'
            vtk_s += 'DATASET UNSTRUCTURED_GRID\n'
            if len(self.points[0]) == 2:
                # Points
                vtk_s += 'POINTS {} float\n'.format(len(self.points))
                for pt in self.points:
                    vtk_s += ' '.join(['{: e}'.format(x) for x in pt]) + ' 0\n'
                vtk_s += '\n'

                # Cells
                n_cells = len(self.regions)
                n_data_total = 0
                cells = 'CELLS {}'.format(n_cells) + ' {}\n'
                pts = np.array(self.points)
                for facets in self.regions:
                    vloop = kp_loop([self.facets[f] for f in facets])
                    n_kp = len(vloop)

                    v1 = pts[vloop[1]] - pts[vloop[0]]
                    v2 = pts[vloop[2]] - pts[vloop[0]]
                    cross_p = np.cross(v1, v2)

                    cells += '{} '.format(n_kp)
                    if cross_p > 0:
                        cells += ' '.join([str(kp) for kp in vloop])
                    else:
                        cells += ' '.join([str(kp) for kp in vloop[::-1]])
                    cells += '\n'
                    n_data_total += 1 + n_kp
                vtk_s += cells.format(n_data_total)

                # Cell Types
                vtk_s += 'CELL_TYPES {}\n'.format(n_cells)
                vtk_s += ''.join(n_cells * ['7\n'])

            else:
                # Points
                vtk_s += 'POINTS {} float\n'.format(len(self.points))
                for pt in self.points:
                    vtk_s += ' '.join(['{: e}'.format(x) for x in pt]) + '\n'
                vtk_s += '\n'

                # Cells
                n_cells = len(self.regions)
                cells = 'CELLS {} '.format(n_cells) + '{}\n'
                n_data_total = 0
                pts = np.array(self.points)
                for facets in self.regions:
                    # Get region center
                    kps = list({kp for f in facets for kp in self.facets[f]})
                    cen = pts[kps].mean(axis=0)  # estimate of center

                    # Write facets
                    n_data_region = 1
                    line = '{} ' + str(len(facets))
                    for f_num in facets:
                        facet = self.facets[f_num]
                        f_len = len(facet)

                        # Determine clockwise or counter-clockwise
                        v1 = pts[facet[1]] - pts[facet[0]]
                        v2 = pts[facet[2]] - pts[facet[0]]
                        norm_vec = np.cross(v1, v2)
                        cen_rel = cen - pts[facet[0]]
                        dot_p = np.dot(norm_vec, cen_rel)

                        line += ' {} '.format(f_len)
                        if dot_p < 0:
                            line += ' '.join([str(kp) for kp in facet])
                        else:
                            line += ' '.join([str(kp) for kp in facet[::-1]])
                        n_data_region += 1 + f_len
                    line += '\n'
                    cells += line.format(n_data_region)
                    n_data_total += 1 + n_data_region
                vtk_s += cells.format(n_data_total)
                vtk_s += '\n'

                # Cell Types
                vtk_s += 'CELL_TYPES {}\n'.format(n_cells)
                vtk_s += ''.join(n_cells * ['42\n'])

            # Cell Data
            vtk_s += '\nCELL_DATA ' + str(n_cells) + '\n'
            vtk_s += 'SCALARS seed int 1 \n'
            vtk_s += 'LOOKUP_TABLE seed\n'
            vtk_s += ''.join([str(a) + '\n' for a in self.seed_numbers])

            vtk_s += 'SCALARS phase int 1 \n'
            vtk_s += 'LOOKUP_TABLE phase\n'
            vtk_s += ''.join([str(a) + '\n' for a in self.phase_numbers])

            vtk_s += 'SCALARS volume float 1 \n'
            vtk_s += 'LOOKUP_TABLE volume\n'
            vtk_s += ''.join([str(a) + '\n' for a in self.volumes])

            with open(filename, 'w') as file:
                file.write(vtk_s)

        else:
            e_str = 'Cannot understand format string ' + str(format) + '.'
            raise ValueError(e_str)

    @classmethod
    def from_file(cls, filename):
        """Read PolyMesh from file.

        This function reads in a polygon mesh from a file and creates an
        instance from that file. Currently the only supported file type
        is the output from :meth:`.write` with the ``format='txt'`` option.

        Args:
            filename (str): Name of file to read from.

        Returns:
            PolyMesh: The instance of the class written to the file.

        """
        with open(filename, 'r') as file:
            stage = 0
            pts = []
            facets = []
            f_neighbors = []
            regions = []
            seed_numbers = []
            phase_numbers = []
            volumes = []
            for line in file.readlines():
                if 'Mesh Points'.lower() in line.lower():
                    n_pts = int(line.split(':')[1])
                    stage = 'points'
                elif 'Mesh Facets'.lower() in line.lower():
                    n_fts = int(line.split(':')[1])
                    stage = 'facets'
                elif 'Facet Neighbors'.lower() in line.lower():
                    n_nns = int(line.split(':')[1])
                    stage = 'facet neighbors'
                elif 'Mesh Regions'.lower() in line.lower():
                    n_rns = int(line.split(':')[1])
                    stage = 'regions'
                elif 'Seed Numbers'.lower() in line.lower():
                    n_sns = int(line.split(':')[1])
                    stage = 'seed numbers'
                elif 'Phase Numbers'.lower() in line.lower():
                    n_pns = int(line.split(':')[1])
                    stage = 'phase numbers'
                elif 'Volumes'.lower() in line.lower():
                    n_vols = int(line.split(':')[1])
                    stage = 'volumes'
                else:
                    if stage == 'points':
                        pts.append([float(x) for x in line.split(',')])
                    elif stage == 'facets':
                        facets.append([int(kp) for kp in line.split(',')])
                    elif stage == 'facet neighbors':
                        f_neighbors.append([int(n) for n in line.split(',')])
                    elif stage == 'regions':
                        regions.append([int(f) for f in line.split(',')])
                    elif stage == 'seed numbers':
                        seed_numbers.append(_misc.from_str(line))
                    elif stage == 'phase numbers':
                        phase_numbers.append(_misc.from_str(line))
                    elif stage == 'volumes':
                        volumes.append(_misc.from_str(line))
                    else:
                        pass

        # check the inputs
        assert len(pts) == n_pts
        assert len(facets) == n_fts
        assert len(regions) == n_rns
        assert len(seed_numbers) == n_sns
        assert len(phase_numbers) == n_pns
        assert len(volumes) == n_vols

        if len(f_neighbors) == 0:
            f_neighbors = None
        else:
            assert len(f_neighbors) == n_nns

        return cls(pts, facets, regions, seed_numbers, phase_numbers,
                   volumes=volumes, facet_neighbors=f_neighbors)

    # ----------------------------------------------------------------------- #
    # Construct from Seed List                                                #
    # ----------------------------------------------------------------------- #
    @classmethod
    def from_seeds(cls, seedlist, domain, edge_opt=False, n_iter=100,
                   verbose=False):
        """Create from :class:`.SeedList` and a domain.

        This function creates a polygon/polyhedron mesh from a seed list and
        a domain. It relies on the pyvoro package, which wraps `Voro++`_.
        The mesh is a Voronoi power diagram / Laguerre tessellationself.

        The pyvoro package operates on rectangular domains, so other domains
        are meshed in 2D by meshing in a bounding box then the boundary cells
        are clipped to the domain boundary.
        Currently non-rectangular domains in 3D are not supported.

        This function also includes the option to maximize the shortest edges
        in the polygonal/polyhedral mesh. Short edges cause numerical
        issues in finite element analysis - setting `edge_opt` to True can
        improve mesh quality with minimal changes to the microstructure.

        Args:
            seedlist (SeedList): A list of seeds in the microstructure.
            domain (from :mod:`microstructpy.geometry`): The domain to be
                filled by the seed.
            edge_opt (bool): *(optional)* This option will maximize the minimum
                edge length in the PolyMesh. The seeds associated with the
                shortest edge are displaced randomly to find improvement and
                this process iterates until `n_iter` attempts have been made
                for a given edge. Defaults to False.
            n_iter (int): *(optional)* Maximum number of iterations per edge
                during optimization. Ignored if `edge_opt` set to False.
                Defaults to 100.
            verbose (bool): *(optional)* Print status of edge optimization to
                screen. Defaults to False.

        Returns:
            PolyMesh: A polygon/polyhedron mesh.

        .. _`Voro++`: http://math.lbl.gov/voro++/

        """
        # Collect all breakdowns
        bkdwn2seed = np.array([], dtype='int')
        bkdwns = np.array([])
        for seed_num, seed in enumerate(seedlist):
            if len(seed.breakdown) == 0:
                seed.update_breakdown()
            bkdwn = np.array(seed.breakdown).reshape(-1, domain.n_dim + 1)
            in_mask = domain.within(bkdwn[:, :-1])
            breakdown = bkdwn[in_mask]

            m, n = breakdown.shape
            bkdwns = np.concatenate((bkdwns.reshape(-1, n), breakdown))
            bkdwn2seed = np.append(bkdwn2seed, np.full(m, seed_num))

        n_pts = bkdwns.shape[0]
        n_dim = bkdwns.shape[1] - 1

        # modify point list and boundaries if necessary
        geom = {2: geometry.Rectangle, 3: geometry.Box}
        voro_dom = geom[n_dim](limits=domain.limits)

        # get domain limits
        lims = voro_dom.limits

        # clip points from voro domain
        flag_val = min(bkdwn2seed) - 1
        n_pad = bkdwns.shape[0] - n_pts
        bkdwn2seed = np.pad(bkdwn2seed, (0, n_pad), 'constant',
                            constant_values=(0, flag_val))

        cens = bkdwns[:, :-1]
        rads = bkdwns[:, -1]

        # get block size
        sz = 2 * max(rads)
        if np.isclose(sz, 0):
            sz = 0.1 * np.min([ub - lb for lb, ub in lims])

        # remove extraneous breakdowns
        removing_pts = True
        while removing_pts:
            # Create a temporary file to run pyvoro
            call_str = 'import pyvoro\n\n'

            call_str += 'pts = ['
            beg_str = ',\n' + len('pts = [') * ' '
            call_str += beg_str.join([str(np.array(p).tolist()) for p in cens])
            call_str += ']\n\n'

            call_str += 'lims = ' + str(np.array(lims).tolist()) + '\n\n'

            call_str += 'sz = ' + str(sz) + '\n\n'

            call_str += 'rads = ['
            beg_str = ',\n' + len('rads = [') * ' '
            call_str += beg_str.join([str(rad) for rad in rads])
            call_str += ']\n\n'

            call_str += 'pyvoro.compute_'
            if n_dim == 2:
                call_str += '2d_'
            call_str += 'voronoi(pts, lims, sz, rads)\n'

            file = tempfile.NamedTemporaryFile(mode='w', suffix='.py',
                                               delete=False)
            file.write(call_str)
            call_filename = file.name
            file.close()

            # Run pyvoro
            p = subprocess.Popen([sys.executable, call_filename],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            p_out, _ = p.communicate()
            try:
                p.terminate()
            except OSError:
                pass

            os.remove(call_filename)

            # if there is output, remove those cells from the list
            out_str = p_out.decode('utf-8')
            if out_str:
                inds = [int(s) for s in out_str.split(':')[-1].split()]
                mask = np.full(len(rads), True)
                mask[inds] = False
                cens = cens[mask]
                rads = rads[mask]
                bkdwn2seed = bkdwn2seed[mask]
            else:
                removing_pts = False

        missing_seeds = set(range(len(seedlist))) - set(bkdwn2seed)
        assert not missing_seeds, str(missing_seeds)
        # compute voronoi diagram
        voro_fun = {2: pyvoro.compute_2d_voronoi,
                    3: pyvoro.compute_voronoi}[n_dim]
        voro = voro_fun(cens, lims, sz, rads)

        # Get only the cells within the domain
        cell_mask = np.full(len(bkdwn2seed), True, dtype='bool')
        rect_doms = ['square', 'cube', 'rectangle', 'box', 'nbox']
        if type(domain).__name__.lower() not in rect_doms:
            for cell_num, cell in enumerate(voro):
                cell_pts = np.array(cell['vertices'])
                cell_mask[cell_num] = np.any(domain.within(cell_pts))
        bkdwn2seed = bkdwn2seed[cell_mask]

        new_cell_nums = np.full(len(cell_mask), -1, dtype='int')
        new_cell_nums[cell_mask] = np.arange(np.sum(cell_mask))

        reduced_voro = []
        for old_cell_num, cell in enumerate(voro):
            # update the numbers of adjacent cells
            faces = cell['faces']
            for face in faces:
                old_adj_cell_num = face['adjacent_cell']
                if old_adj_cell_num >= 0:
                    new_adj_cell_num = new_cell_nums[old_adj_cell_num]
                    face['adjacent_cell'] = new_adj_cell_num
            cell['faces'] = faces

            # add cell to voro
            if cell_mask[old_cell_num]:
                reduced_voro.append(cell)

        # Clip cells to domain
        voro = [_clip_cell(c, domain) for c in reduced_voro]

        # create global key point and facet lists
        pts_global = []
        pts_conn = []
        local_kp_conn = {}

        for cell_num, cell_data in enumerate(voro):
            pts_local = cell_data['vertices']

            for face_data in cell_data['faces']:
                adj_cell = face_data['adjacent_cell']
                simplex_local = face_data['vertices']
                for kp_local in simplex_local:
                    key = (cell_num, kp_local)
                    if key in local_kp_conn:
                        kp_global = local_kp_conn[key]

                    else:
                        kp_global = len(pts_global)
                        pt = pts_local[kp_local]
                        pts_global.append(pt)
                        pts_conn.append({cell_num: kp_local})
                        local_kp_conn[key] = kp_global

                    if (adj_cell >= 0) and (adj_cell < len(voro)):
                        conn_info = pts_conn[kp_global]
                        if adj_cell not in conn_info:
                            adj_cell_data = voro[adj_cell]
                            adj_pts_local = np.array(adj_cell_data['vertices'])
                            rel_pos = adj_pts_local - pts_global[kp_global]
                            sq_dist = np.sum(rel_pos * rel_pos, axis=-1)
                            adj_kp_local = np.argmin(sq_dist)

                            adj_key = (adj_cell, adj_kp_local)
                            local_kp_conn[adj_key] = kp_global
                            pts_conn[kp_global][adj_cell] = adj_kp_local

        # create facet and region lists
        facet_list = []
        facet_neighbor_list = []
        region_list = [[] for cell in voro]
        for cell_num, cell_data in enumerate(voro):
            for face_data in cell_data['faces']:
                adj_cell_num = face_data['adjacent_cell']
                if adj_cell_num >= len(voro):
                    adj_cell_num = -1

                if adj_cell_num < cell_num:
                    neighbor_pair = (adj_cell_num, cell_num)
                    s_lcl = face_data['vertices']
                    s_glbl = [local_kp_conn[(cell_num, kp)] for kp in s_lcl]
                    if adj_cell_num < 0:
                        pts_f = [pts_global[kp] for kp in s_glbl]
                        if not _is_outward(pts_f, adj_cell_num):
                            s_glbl.reverse()

                    face_num = len(facet_list)
                    facet_list.append(s_glbl)
                    facet_neighbor_list.append(neighbor_pair)

                    for f_cell_num in neighbor_pair:
                        if (f_cell_num >= 0) and (f_cell_num < len(voro)):
                            region_list[f_cell_num].append(face_num)

        # create phase number list
        phase_nums = [seedlist[i].phase for i in bkdwn2seed]

        # Create volume list
        vols = [cell['volume'] for cell in voro]

        # Create initial mesh
        pmesh = cls(pts_global, facet_list, region_list, bkdwn2seed,
                    phase_nums, facet_neighbor_list, vols)

        # short edge optimization
        if edge_opt:
            seed2bkdwn = {i: [] for i in range(len(seedlist))}
            for i, n in enumerate(pmesh.seed_numbers):
                seed2bkdwn[n].append(i)

            # Find the shorted edge
            edge_lens = _edge_lengths(pmesh)
            min_edge = _shortest_edge(edge_lens)
            min_len = edge_lens[min_edge]['length']

            # Format verbose print string
            n_kps = len(pmesh.points)
            n_kp_space = int(np.log10(n_kps)) + 1
            n_iter_space = int(np.log10(n_iter))
            v_fmt = 'min length: {0:.3e} | '
            v_fmt += 'edge: {1[0]:' + str(n_kp_space) + 'd}, '
            v_fmt += '{1[1]:' + str(n_kp_space) + 'd} | '
            v_fmt += 'n iter: {2:' + str(n_iter_space) + 'd} / '
            v_fmt += str(n_iter)

            i_n_attempts = 0
            while i_n_attempts < n_iter:
                print(v_fmt.format(min_len, min_edge, i_n_attempts))
                # Create Displacement
                max_step_size = float('inf')
                step_fracs = 2 * np.random.rand(3) - 1  # [-1, 1]
                new_cens = np.copy(cens)

                e_neighs = edge_lens[min_edge]['regions']
                edge_pts = np.array(pmesh.points)[list(min_edge)]
                for region_num in e_neighs:
                    if region_num >= 0:
                        step_size = 0.1 * rads[region_num]
                        max_step_size = min(max_step_size, step_size)
                for f, region_num in zip(step_fracs, e_neighs):
                    if region_num >= 0:
                        e_norm_vec = _point_line_vec(cens[region_num],
                                                     edge_pts)
                        step = f * max_step_size * e_norm_vec
                        new_cens[region_num] += step

                # Update Seeds
                new_bkdwns = [list(c) + [r] for c, r in zip(new_cens, rads)]
                for i, seed in enumerate(seedlist):
                    seed.breakdown = [new_bkdwns[j] for j in seed2bkdwn[i]]

                # Create New Polygonal Mesh
                try:
                    new_pmesh = cls.from_seeds(seedlist, domain,
                                               edge_opt=False)
                except AssertionError:
                    i_n_attempts += 1
                    continue

                new_edge_lens = _edge_lengths(new_pmesh)
                new_min_edge = _shortest_edge(new_edge_lens)
                new_min_len = new_edge_lens[new_min_edge]['length']

                if new_min_len > min_len:
                    if new_min_edge != min_edge:
                        i_n_attempts = 0
                    else:
                        i_n_attempts += 1
                    edge_lens = new_edge_lens
                    pmesh = new_pmesh
                    min_len = new_min_len
                    min_edge = new_min_edge
                    cens = new_cens
                else:
                    i_n_attempts += 1
        return pmesh

    # ----------------------------------------------------------------------- #
    # Plot Mesh                                                               #
    # ----------------------------------------------------------------------- #
    def plot(self, index_by='seed', material=[], loc=0, **kwargs):
        """Plot the mesh.

        This function plots the polygon mesh.
        In 2D, this creates a class:`matplotlib.collections.PolyCollection`
        and adds it to the current axes.
        In 3D, it creates a
        :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection` and
        adds it to the current axes.
        The keyword arguments are passed though to matplotlib.

        Args:
            index_by (str): *(optional)* {'facet' | 'material' | 'seed'}
                Flag for indexing into the other arrays passed into the
                function. For example,
                ``plot(index_by='material', color=['blue', 'red'])`` will plot
                the regions with ``phase_number`` equal to 0 in blue, and
                regions with ``phase_number`` equal to 1 in red. The facet
                option is only available for 3D plots. Defaults to 'seed'.
            material (list): *(optional)* Names of material phases. One entry
                per material phase (the ``index_by`` argument is ignored).
                If this argument is set, a legend is added to the plot with
                one entry per material.
            loc (int or str): *(optional)* The location of the legend,
                if 'material' is specified. This argument is passed directly
                through to :func:`matplotlib.pyplot.legend`. Defaults to 0,
                which is 'best' in matplotlib.
            **kwargs: Keyword arguments for matplotlib.

        """
        n_dim = len(self.points[0])
        if n_dim == 2:
            ax = plt.gca()
        else:
            ax = plt.gcf().gca(projection=Axes3D.name)
        n_obj = _misc.ax_objects(ax)
        if n_obj > 0:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        else:
            xlim = [float('inf'), -float('inf')]
            ylim = [float('inf'), -float('inf')]
        if n_dim == 2:
            # create vertex loops for each poly
            vloops = [kp_loop([self.facets[f] for f in r]) for r in
                      self.regions]
            # create poly input
            xy = [np.array([self.points[kp] for kp in lp]) for lp in vloops]

            plt_kwargs = {}
            for key, value in kwargs.items():
                if type(value) in (list, np.array):
                    plt_value = []
                    for s, p in zip(self.seed_numbers, self.phase_numbers):
                        if index_by == 'material':
                            region_value = value[p]
                        elif index_by == 'seed':
                            region_value = value[s]
                        else:
                            e_str = 'Cannot index by {}.'.format(index_by)
                            raise ValueError(e_str)
                        plt_value.append(region_value)
                else:
                    plt_value = value
                plt_kwargs[key] = plt_value
            pc = collections.PolyCollection(xy, **plt_kwargs)
            ax.add_collection(pc)
            ax.autoscale_view()
        elif n_dim == 3:
            if n_obj > 0:
                zlim = ax.get_zlim()
            else:
                zlim = [float('inf'), -float('inf')]
            self.plot_facets(index_by=index_by, **kwargs)

        else:
            raise NotImplementedError('Cannot plot in ' + str(n_dim) + 'D.')

        # Add legend
        if material and index_by in ('seed', 'material'):
            p_kwargs = [{'label': m} for m in material]
            s2p = {s: p for s, p in zip(self.seed_numbers, self.phase_numbers)}
            for key, value in kwargs.items():
                if type(value) in (list, np.array):
                    if index_by == 'material':
                        for p, v in enumerate(value):
                            p_kwargs[p][key] = v
                    else:
                        for s, v in enumerate(value):
                            p = s2p[s]
                            p_kwargs[p][key] = v
                else:
                    for i, m in enumerate(material):
                        p_kwargs[i][key] = value

            # Replace plural keywords
            for p_kw in p_kwargs:
                for kw in _misc.mpl_plural_kwargs:
                    if kw in p_kw:
                        p_kw[kw[:-1]] = p_kw[kw]
                        del p_kw[kw]
            handles = [patches.Patch(**p_kw) for p_kw in p_kwargs]
            ax.legend(handles=handles, loc=loc)

        # Adjust Axes
        mins = np.array(self.points).min(axis=0)
        maxs = np.array(self.points).max(axis=0)
        xlim = (min(xlim[0], mins[0]), max(xlim[1], maxs[0]))
        ylim = (min(ylim[0], mins[1]), max(ylim[1], maxs[1]))
        if n_dim == 2:
            plt.axis('square')
            plt.xlim(xlim)
            plt.ylim(ylim)
        elif n_dim == 3:
            zlim = (min(zlim[0], mins[2]), max(zlim[1], maxs[2]))
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)

            _misc.axisEqual3D(ax)

    def plot_facets(self, index_by='seed', hide_interior=True, **kwargs):
        """Plot PolyMesh facets.

        This function plots the facets of the polygon mesh, rather than the
        regions.
        In 2D, it adds a :class:`matplotlib.collections.LineCollection` to the
        current axes.
        In 3D, it adds a
        :class:`mpl_toolkits.mplot3d.art3d.Poly3DCollection`
        with ``facecolors='none'``.
        The keyword arguments are passed though to matplotlib.

        Args:
            index_by (str): *(optional)* {'facet' | 'material' | 'seed'}
                Flag for indexing into the other arrays passed into the
                function. For example,
                ``plot(index_by='material', color=['blue', 'red'])`` will plot
                the regions with ``phase_number`` equal to 0 in blue, and
                regions with ``phase`` equal to 1 in red. The facet option is
                only available for 3D plots. Defaults to 'seed'.
            hide_interior (bool): If True, removes interior facets from the
                output plot. This avoids occasional matplotlib issue where
                interior facets are shown in output plots.
            **kwargs (dict): Keyword arguments for matplotlib.

        """
        f_kwargs = {}
        for key, value in kwargs.items():
            if type(value) in (list, np.array):
                f_values = []
                for fn in range(len(self.facets)):
                    neighs = self.facet_neighbors[fn]
                    r = max(neighs)
                    sn = self.seed_numbers[r]
                    pn = self.phase_numbers[r]
                    if index_by == 'facet':
                        ind = fn
                    elif index_by == 'material':
                        ind = pn
                    elif index_by == 'seed':
                        ind = sn
                    else:
                        e_str = 'Cannot index by {}.'.format(index_by)
                        raise ValueError(e_str)
                    v = value[ind]
                    f_values.append(v)
                f_kwargs[key] = f_values
            else:
                f_kwargs[key] = value

        n_dim = len(self.points[0])
        if n_dim == 2:
            ax = plt.gca()
        else:
            ax = plt.gcf().gca(projection=Axes3D.name)
        n_obj = _misc.ax_objects(ax)
        if n_obj > 0:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        else:
            xlim = [float('inf'), -float('inf')]
            ylim = [float('inf'), -float('inf')]
        if n_dim == 2:
            xy = [np.array([self.points[kp] for kp in f]) for f in self.facets]

            pc = collections.LineCollection(xy, **f_kwargs)
            ax.add_collection(pc)
            ax.autoscale_view()
        else:

            if ax.has_data:
                zlim = ax.get_zlim()
            else:
                zlim = [float('inf'), -float('inf')]

            if hide_interior:
                f_mask = [min(fn) < 0 for fn in self.facet_neighbors]
                xy = [np.array([self.points[kp] for kp in f]) for m, f in
                      zip(f_mask, self.facets) if m]
                list_kws = [k for k, vl in f_kwargs.items()
                            if isinstance(vl, list)]
                plt_kwargs = {k: vl for k, vl in f_kwargs.items() if
                              k not in list_kws}
                for k in list_kws:
                    v = [val for val, m in zip(f_kwargs[k], f_mask) if m]
                    plt_kwargs[k] = v
            else:
                xy = [np.array([self.points[kp] for kp in f]) for f in
                      self.facets]
                plt_kwargs = f_kwargs
            pc = Poly3DCollection(xy, **plt_kwargs)
            ax.add_collection(pc)

        # Adjust Axes
        mins = np.array(self.points).min(axis=0)
        maxs = np.array(self.points).max(axis=0)

        xlim = (min(xlim[0], mins[0]), max(xlim[1], maxs[0]))
        ylim = (min(ylim[0], mins[1]), max(ylim[1], maxs[1]))
        if n_dim == 2:
            plt.axis('square')
            plt.xlim(xlim)
            plt.ylim(ylim)
        if n_dim == 3:
            zlim = (min(zlim[0], mins[2]), max(zlim[1], maxs[2]))

            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.set_zlim(zlim)
            _misc.axisEqual3D(ax)

    # ----------------------------------------------------------------------- #
    # Mesh Equality                                                           #
    # ----------------------------------------------------------------------- #
    def __eq__(self, other_mesh):
        # check type
        if type(other_mesh) is not PolyMesh:
            print('not same type')
            return False

        # check that the lengths are all the same
        same = True
        same &= len(self.points) == len(other_mesh.points)
        same &= len(self.facets) == len(other_mesh.facets)
        same &= len(self.regions) == len(other_mesh.regions)
        same &= len(self.seed_numbers) == len(other_mesh.seed_numbers)
        same &= len(self.phase_numbers) == len(other_mesh.phase_numbers)
        if not same:
            print('not same length')
            return False

        # check that the vertices have the same coordinates
        pt_dists = distance.cdist(self.points, other_mesh.points)
        same_pt = np.isclose(pt_dists, 0)
        same_ints = same_pt.astype(int)
        same &= np.all(same_ints.sum(axis=0) == 1)
        same &= np.all(same_ints.sum(axis=1) == 1)
        if not same:
            print('not same verts')
            return False

        kp_conv = np.argwhere(same_pt)
        kp_other = kp_conv[:, 1]
        print('transform')
        print(np.array(kp_other))

        # check that the facets are the same
        facets_in_other_kps = [[kp_other[kp] for kp in f] for f in self.facets]
        o_fnum = []
        for i, s_facet in enumerate(facets_in_other_kps):
            for j, o_facet in enumerate(other_mesh.facets):
                if j in o_fnum:
                    continue
                else:
                    if set(s_facet) == set(o_facet):
                        o_fnum.append(j)
                        break

            if len(o_fnum) != i + 1:
                print('not same facets')
                return False

        # check that the regions are the same
        regions_in_other_fnums = [[o_fnum[f] for f in r] for r in self.regions]
        o_rnum = []
        for i, s_region in enumerate(regions_in_other_fnums):
            for j, o_region in enumerate(other_mesh.regions):
                if j in o_rnum:
                    continue
                else:
                    if set(s_region) == set(o_region):
                        o_rnum.append(j)
                        break

            if len(o_rnum) != i + 1:
                print('not same regions')
                return False

        # check that the seed numbers are the same
        s_seed_nums = np.array(self.seed_numbers)
        o_seed_nums = np.array(other_mesh.seed_numbers)
        same &= np.all(s_seed_nums == o_seed_nums[o_rnum])
        print('checking seed numbers', same)

        # check that the phase numbers are the same
        s_phase_nums = np.array(self.phase_numbers)
        o_phase_nums = np.array(other_mesh.phase_numbers)
        same &= np.all(s_phase_nums == o_phase_nums[o_rnum])
        print('checking phase numbers', same)

        return same


def kp_loop(kp_pairs):
    loop = list(kp_pairs[0])
    kp_arr = np.array(kp_pairs[1:])
    while kp_arr.shape[0] > 0:
        kp_find = loop[-1]
        has_kp = np.any(kp_arr == kp_find, axis=1)
        row = kp_arr[has_kp]

        loop.append(row[row != kp_find][0])
        kp_arr = kp_arr[~has_kp]
    assert loop[0] == loop[-1]
    return loop[:-1]


def _clip_cell(cell_data, domain):
    domain_name = type(domain).__name__.lower()
    if domain_name in ['rectangle', 'square', 'box', 'cube']:
        return cell_data

    if domain.n_dim == 2:
        pts = np.array(cell_data['vertices'])
        if np.all(domain.within(pts)):
            return cell_data

        # split the edges that contain the boundary
        new_adj = np.copy(cell_data['adjacency'])
        new_faces = []
        new_pts = np.copy(cell_data['vertices'])
        new_kps = []

        for face in cell_data['faces']:
            adj_cell = face['adjacent_cell']
            verts = face['vertices']
            face_pts = pts[verts]
            pts_within = domain.within(face_pts)
            if np.all(pts_within) or np.all(~pts_within):
                new_faces.append(face)
                continue
            crossing_pt = _segment_cross(face_pts, domain)

            # Add point to list of vertices and face to list of faces
            crossing_kp = len(new_pts)
            new_pts = np.vstack((new_pts, crossing_pt.reshape(1, -1)))
            new_kps.append(crossing_kp)

            for kp_i, kp in enumerate(verts):
                kp_other = verts[1 - kp_i]
                new_adj[kp] = [kp_other, crossing_kp]

                new_verts = [kp, crossing_kp]
                new_faces.append({'adjacent_cell': adj_cell,
                                  'vertices': new_verts})

        # add divider face
        new_faces.append({'adjacent_cell': -1, 'vertices': new_kps})

        # Create cell within the domain
        new_within = domain.within(new_pts)
        new_within[new_kps] = True

        within_pts = new_pts[new_within]
        kp_conv = np.full(len(new_pts), -1, dtype='int')
        kp_conv[new_within] = np.arange(np.sum(new_within))

        within_adj = [[] for pt in within_pts]
        within_faces = []
        for face in new_faces:
            adj_cell = face['adjacent_cell']
            old_verts = face['vertices']
            new_verts = [kp_conv[v] for v in old_verts]
            within_face = {'adjacent_cell': adj_cell, 'vertices': new_verts}
            if all([v >= 0 for v in new_verts]):
                within_faces.append(within_face)
                within_adj[new_verts[0]].append(new_verts[1])
                within_adj[new_verts[1]].append(new_verts[0])

        # Compute cell area
        within_loop = kp_loop([f['vertices'] for f in within_faces])
        within_area = _loop_area(within_pts, within_loop)

        new_cell_data = {'adjacency': within_adj,
                         'faces': within_faces,
                         'original': cell_data['original'],
                         'vertices': within_pts,
                         'volume': within_area}

        return new_cell_data

    w_str = 'Cannot clip cells to fit to a ' + domain_name + '.'
    w_str = ' Currently 3D geometries are not supported, other than boxes.'
    warnings.warn(w_str, RuntimeWarning)
    return cell_data


def _segment_cross(pts, domain):
    end_pts = np.copy(pts)
    ds = np.inf
    while ds > 1e-12:
        within = domain.within(end_pts)
        pt = end_pts.mean(axis=0)
        pt_within = domain.within(pt)
        if within[0] == pt_within:
            end_pts[0] = pt
        else:
            end_pts[1] = pt
        dx = end_pts[1] - end_pts[0]
        ds = np.linalg.norm(dx)
    return pt


def _loop_area(pts, loop):
    double_area = 0

    n = len(loop)
    for i in range(n):
        ip1 = (i + 1) % n

        xi = pts[i][0]
        yi = pts[i][1]
        xip1 = pts[ip1][0]
        yip1 = pts[ip1][1]

        det = xi * yip1 - xip1 * yi
        double_area += det
    return 0.5 * np.abs(double_area)


def _is_outward(pt_list, voropp_face_num):
    n_dim = len(pt_list[0])

    voropp_sgn = 1-2*(voropp_face_num % 2)
    voropp_axis = int((-voropp_face_num - 1)/2)
    face_vec = voropp_sgn * np.eye(n_dim)[voropp_axis]

    if n_dim == 2:
        pt1 = pt_list[0]
        pt2 = pt_list[1]
        rel_pos = np.array(pt2) - np.array(pt1)
        n_vec = np.array([-rel_pos[1], rel_pos[0]])
    elif n_dim == 3:
        pt1 = pt_list[0]
        pt2 = pt_list[1]
        pt3 = pt_list[2]
        r1 = np.array(pt2) - np.array(pt1)
        r2 = np.array(pt3) - np.array(pt1)
        n_vec = np.cross(r1, r2)
    else:
        raise ValueError('Function does not support {}D.'.format(n_dim))

    n_u = n_vec / np.linalg.norm(n_vec)
    return np.dot(n_u, face_vec) > 0


def _edge_lengths(pmesh):
    edge_lens = {}  # (kp1, kp2): {'length': #, 'regions': set()}
    for i, f in enumerate(pmesh.facets):
        n = len(f)
        facet_kp_pairs = [(f[k], f[(k + 1) % n]) for k in range(n)]
        for pair in facet_kp_pairs:
            key = tuple(sorted(pair))
            if key not in edge_lens:  # calculate edge length
                pt1 = pmesh.points[key[0]]
                pt2 = pmesh.points[key[1]]
                rel_pos = np.array(pt2) - np.array(pt1)
                edge_len = np.linalg.norm(rel_pos)
                edge_lens[key] = {
                    'length': edge_len,
                    'regions': set(),
                    }
            neighs = pmesh.facet_neighbors[i]
            edge_lens[key]['regions'] |= set(neighs)
    return edge_lens


def _shortest_edge(edge_lens):
    min_len = float('inf')
    min_pair = (-1, -1)
    for pair in edge_lens:
        length = edge_lens[pair]['length']
        if length < min_len:
            min_len = length
            min_pair = pair
    return min_pair


def _point_line_vec(pt, line_pts):
    ptA, ptB = line_pts
    n_vec = (ptB - ptA) / np.linalg.norm(ptB - ptA)

    rel_pos = ptA - pt
    proj = np.dot(rel_pos, n_vec) * n_vec
    dist_vec = rel_pos - proj

    u_vec = dist_vec / np.linalg.norm(dist_vec)
    return u_vec



from microstructpy.geometry.box import Box
from microstructpy.geometry.box import Cube
from microstructpy.geometry.circle import Circle
from microstructpy.geometry.ellipse import Ellipse
from microstructpy.geometry.ellipsoid import Ellipsoid
from microstructpy.geometry.rectangle import Rectangle
from microstructpy.geometry.rectangle import Square
from microstructpy.geometry.sphere import Sphere

__all__ = ['Box', 'Cube', 'Circle', 'Ellipse', 'Ellipsoid',
           'Rectangle', 'Square', 'Sphere']


def factory(name, **kwargs):
    """Factory method for geometries.

    This function returns a geometry based on a string containing the
    name of the geometry and keyword arguments defining the geometry.

    .. note::

        The function call is ``factory(name, **kwargs)``. Sphinx autodocs
        has dropped the first parameter.

    Args:
        name (str): {'box' | 'cube' | 'ellipse' | 'ellipsoid' | 'circle' | 
            'rectangle' | 'square' | 'sphere'} Name of geometry.
        **kwargs (dict): Arguments defining the geometry.

    """
    geom = name.strip().lower()
    if geom in ('box', 'rectangular prism', 'cuboid'):
        return Box(**kwargs)
    elif geom == 'cube':
        return Cube(**kwargs)
    elif geom == 'ellipse':
        return Ellipse(**kwargs)
    elif geom == 'ellipsoid':
        return Ellipsoid(**kwargs)
    elif geom == 'circle':
        return Circle(**kwargs)
    elif geom == 'rectangle':
        return Rectangle(**kwargs)
    elif geom == 'square':
        return Square(**kwargs)
    elif geom == 'sphere':
        return Sphere(**kwargs)
    else:
        e_str = 'Cannot recognize geometry name: ' + geom
        raise ValueError(e_str)
